/*
 * Isomorphism.cpp
 *
 *   Created on: Jun 19, 2014
 *       Author: David Rosca
 */

#include "Isomorphism.h"

#include <exception/CommonException.h>

namespace graph
{

namespace isomorphism
{

static inline std::set<UndirectedEdge> stripNames(const std::set<UndirectedEdge> &s)
{
	std::set<UndirectedEdge> out;
	for (const UndirectedEdge &v : s)
		out.insert(UndirectedEdge(v.getFirstNode(), v.getSecondNode()));
	return out;
}

static inline std::set<DirectedEdge> stripNames(const std::set<DirectedEdge> &s)
{
	std::set<DirectedEdge> out;
	for (const DirectedEdge &v : s)
		out.insert(DirectedEdge(v.getFromNode(), v.getToNode()));
	return out;
}

static bool is_equal(const std::set<UndirectedEdge> &first, const std::set<UndirectedEdge> &second, const std::unordered_map<Node, Node> &mapping)
{
	if (first.size() != second.size())
		return false;

	std::set<UndirectedEdge> frst = stripNames(first);
	std::set<UndirectedEdge> scnd = stripNames(second);

	for (const UndirectedEdge &e : frst) {
		UndirectedEdge other1(mapping.at(e.getFirstNode()), mapping.at(e.getSecondNode()), e.getName());
		UndirectedEdge other2(mapping.at(e.getSecondNode()), mapping.at(e.getFirstNode()), e.getName());
		if (scnd.find(other1) == scnd.end() && scnd.find(other2) == scnd.end())
			return false;
	}

	return true;
}

static bool is_equal(const std::set<DirectedEdge> &first, const std::set<DirectedEdge> &second, const std::unordered_map<Node, Node> &mapping)
{
	if (first.size() != second.size())
		return false;

	std::set<DirectedEdge> frst = stripNames(first);
	std::set<DirectedEdge> scnd = stripNames(second);

	for (const DirectedEdge &e : frst) {
		DirectedEdge other(mapping.at(e.getFromNode()), mapping.at(e.getToNode()), e.getName());
		if (scnd.find(other) == scnd.end())
			return false;
	}

	return true;
}

template<typename T>
static bool impl_r(const T &first, const T &second, const std::vector<Node> &nodes1, const std::vector<Node> &nodes2, std::unordered_map<Node, Node> &mapping)
{
	if (nodes1.empty()) {
		for (const Node &n : first.getNodes()) {
			if (!is_equal(first.neighborEdges(n), second.neighborEdges(mapping.at(n)), mapping)) {
				return false;
			}
		}
		return true;
	}

	std::vector<Node> nds1 = nodes1;
	Node node = nodes1.back();
	nds1.pop_back();

	for (size_t i = 0; i < nodes2.size(); ++i) {
		std::vector<Node> nds2 = nodes2;
		mapping.erase(node);
		mapping.insert({node, nodes2[i]});
		nds2.erase(nds2.begin() + i);
		if (impl_r(first, second, nds1, nds2, mapping))
			return true;
	}

	return false;
}

template<typename T>
static bool impl(const T &first, const T &second)
{
	const std::set<Node> &nodes1 = first.getNodes();
	const std::set<Node> &nodes2 = second.getNodes();

	if (nodes1.size() != nodes2.size())
		return false;

	std::vector<Node> n1;
	std::vector<Node> n2;
	std::unordered_map<Node, Node> mapping;
	std::copy(nodes1.begin(), nodes1.end(), std::back_inserter(n1));
	std::copy(nodes2.begin(), nodes2.end(), std::back_inserter(n2));
	return impl_r(first, second, n1, n2, mapping);
}

bool Isomorphism::isomorphism(const Graph &first, const Graph &second)
{
	return dispatch(first.getData(), second.getData());
}

bool Isomorphism::isomorphism(const DirectedGraph &first, const DirectedGraph &second)
{
	return impl(first, second);
}

bool Isomorphism::isomorphism(const UndirectedGraph &first, const UndirectedGraph &second)
{
	return impl(first, second);
}

} // namespace isomorphism

} // namespace graph
