/*
 * FordFulkerson.cpp
 *
 *   Created on: Mar 28, 2016
 *       Author: Jan Broz
 */

#include "FordFulkerson.h"

#include <unordered_map>
#include <queue>
#include <string>
#include <sstream>
#include <climits>
#include <stdexcept>

#include <exception/CommonException.h>


// the flow needs to be assigned to pair of nodes, not edge
/*  or it can be solved with 2-dimensional map (map of maps)
namespace graph {
namespace maximumflow {
	class NodePair {
	 public:
		Node node1;
		Node node2;
	 public:
		NodePair( const Node & node1, const Node & node2 ) : node1( node1 ), node2( node2 ) {}
		NodePair( const DirectedEdge & edge ) : node1( edge.getFromNode() ), node2( edge.getToNode() ) {}
		std::string toString() const {
			return static_cast<std::string>(node1) + static_cast<std::string>(node2);
		}
	};
}
}

namespace std {
	template<>
	struct hash<graph::maximumflow::NodePair> {
		std::size_t operator()( const graph::maximumflow::NodePair & p ) const {
			return std::hash<std::string>()(p.toString());
		}
	};
}
*/

namespace graph {

namespace maximumflow {

enum State { FRESH, OPEN, CLOSED };
struct Context {
	Capacity capacity;
	Flow flow;
	std::unordered_map<Node,Node> prev;
	std::unordered_map<Node,int> sgn;
	std::unordered_map<Node,int> maxFlowTo;
};

static Capacity getCapacity( const DirectedGraph & graph )
{
	Capacity capacity;
	int edge_cap;

	for (const DirectedEdge & edge : graph.getEdges())
		capacity[ edge.getFromNode() ][ edge.getToNode() ] = 0;

	for (const DirectedEdge & edge : graph.getEdges()) {
		if (capacity[ edge.getFromNode() ][ edge.getToNode() ] != 0)
			throw exception::CommonException("FordFulkerson: graph is not simple");
		try {
			edge_cap = graph.getEdgeValue( edge );
		} catch (const std::out_of_range & ex) { // when this edge has no value assigned
			throw exception::CommonException("FordFulkerson: an edge is missing a capacity");
		}
		capacity[ edge.getFromNode() ][ edge.getToNode() ] = edge_cap;
	}

	return capacity;
}

static bool findPath_dir( const DirectedGraph & graph, Node source, Node sink, Context & ctx )
{
	std::unordered_map<Node,State> state;
	std::queue<Node> open;
	Node actual;

	for (const Node & node : graph.getNodes()) {
		state[ node ] = FRESH;
		ctx.maxFlowTo[ node ] = 0;
	}
	ctx.maxFlowTo[ source ] = INT_MAX;
	ctx.prev[ source ] = source;
	state[ source ] = OPEN;
	open.push( source );

	do {
		actual = open.front(); open.pop();
		state[ actual ] = CLOSED;
		for (const Node & succ : graph.successors( actual )) {
			if (state[ succ ] == FRESH  &&  ctx.flow[ actual ][ succ ] < ctx.capacity[ actual ][ succ ]) {
				state[ succ ] = OPEN;
				open.push( succ );
				ctx.prev[ succ ] = actual;
				ctx.sgn[ succ ] = 1;
				ctx.maxFlowTo[ succ ] = std::min( ctx.maxFlowTo[ actual ], ctx.capacity[ actual ][ succ ] - ctx.flow[ actual ][ succ ]);
			}
		}
		for (const Node & pred : graph.predecessors( actual )) {
			if (state[ pred ] == FRESH  &&  ctx.flow[ pred ][ actual ] > 0) {
				state[ pred ] = OPEN;
				open.push( pred );
				ctx.prev[ pred ] = actual;
				ctx.sgn[ pred ] = -1;
				ctx.maxFlowTo[ pred ] = std::min( ctx.maxFlowTo[ actual ], ctx.flow[ pred ][ actual ]);
			}
		}
	} while (!open.empty() && actual != sink);

	return actual == sink;
}

static bool findPath_undir( const DirectedGraph & graph, Node source, Node sink, Context & ctx )
{
	std::unordered_map<Node,State> state;
	std::queue<Node> open;
	Node actual;

	for (const Node & node : graph.getNodes()) {
		state[ node ] = FRESH;
		ctx.maxFlowTo[ node ] = 0;
	}
	ctx.maxFlowTo[ source ] = INT_MAX;
	ctx.prev[ source ] = source;
	state[ source ] = OPEN;
	open.push( source );

	do {
		actual = open.front(); open.pop();
		state[ actual ] = CLOSED;
		for (const Node & succ : graph.successors( actual )) {
			if (state[ succ ] == FRESH  &&  ctx.flow[ actual ][ succ ] < ctx.capacity[ actual ][ succ ]) {
				state[ succ ] = OPEN;
				open.push( succ );
				ctx.prev[ succ ] = actual;
				ctx.maxFlowTo[ succ ] = std::min( ctx.maxFlowTo[ actual ], ctx.capacity[ actual ][ succ ] - ctx.flow[ actual ][ succ ]);
			}
		}
	} while (!open.empty() && actual != sink);

	return actual == sink;
}

static void updateFlow_dir( Node source, Node sink, Context & ctx )
{
/* alternative for finding the increase of flow
		int path_flow = INT_MAX;
        for (v=t; v!=s; v=parent[v])
        {
            u = parent[v];
            path_flow = min(path_flow, rGraph[u][v]);
        }
*/
	int pathFlow = ctx.maxFlowTo[ sink ];
	Node actual = sink;
	while (actual != source) {
		Node prev = ctx.prev[ actual ];
		if (ctx.sgn[ actual ] > 0)
			ctx.flow[ prev ][ actual ] += pathFlow;
		else
			ctx.flow[ actual ][ prev ] -= pathFlow;
		actual = prev;
	}
}

static void updateFlow_undir( Node source, Node sink, Context & ctx )
{
	int pathFlow = ctx.maxFlowTo[ sink ];
	Node actual = sink;
	while (actual != source) {
		Node prev = ctx.prev[ actual ];
		ctx.flow[ prev ][ actual ] += pathFlow;
		ctx.flow[ actual ][ prev ] -= pathFlow;
		actual = prev;
	}
}

static Flow fordfulkerson_impl_dir( const DirectedGraph & graph, Node source, Node sink )
{
	Context ctx;

	ctx.capacity = getCapacity( graph );
	for (const DirectedEdge & edge : graph.getEdges()) {
		ctx.flow[ edge.getFromNode() ][ edge.getToNode() ] = 0;
		ctx.flow[ edge.getToNode() ][ edge.getFromNode() ] = 0;
	}

	while (findPath_dir( graph, source, sink, ctx ))
		updateFlow_dir( source, sink, ctx );

	// assign negative flow for the reversed pairs of nodes ?
	for (auto u : ctx.flow)
		for (auto v : u.second)
			if (ctx.flow[ u.first ][ v.first ] != 0)
				ctx.flow[ v.first ][ u.first ] = -ctx.flow[ u.first ][ v.first ];

	return ctx.flow;
}

static Flow fordfulkerson_impl_undir( const UndirectedGraph & ugraph, Node source, Node sink )
{
	AdjacencyListDirectedGraph graph( ugraph );
	Context ctx;

	ctx.capacity = getCapacity( graph );
	for (const DirectedEdge & edge : graph.getEdges()) {
		ctx.flow[ edge.getFromNode() ][ edge.getToNode() ] = 0;
	}

	while (findPath_undir( graph, source, sink, ctx ))
		updateFlow_undir( source, sink, ctx );

	return ctx.flow;
}

Flow FordFulkerson::fordfulkerson( const Graph & graph, Node source, Node sink )
{
	return dispatch( graph.getData(), source, sink );
}

Flow FordFulkerson::fordfulkerson( const DirectedGraph & graph, Node source, Node sink )
{
	return fordfulkerson_impl_dir( graph, source, sink );
}

Flow FordFulkerson::fordfulkerson( const UndirectedGraph & graph, Node source, Node sink )
{
	return fordfulkerson_impl_undir( graph, source, sink );
}

//auto FordFulkersonUndirectedGraph = FordFulkerson::RegistratorWrapper< graph::UndirectedGraph, graph::UndirectedGraph >( FordFulkerson::getInstance(), FordFulkerson::fordfulkerson );

}

}
