/*
 * RegExpDerivation.cpp
 *
 *  Created on: 19. 1. 2014
 *	  Author: Tomas Pecka
 */

#include "RegExpDerivation.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>
#include <regexp/formal/FormalRegExpElements.h>

#include "../properties/RegExpEpsilon.h"

namespace regexp
{

regexp::RegExp RegExpDerivation::derivation(const regexp::RegExp& regexp, const string::LinearString& string) {
	return dispatch(regexp.getData(), string);
}

regexp::FormalRegExp RegExpDerivation::derivation(const regexp::FormalRegExp& regexp, const string::LinearString& string) {
	std::unique_ptr < regexp::FormalRegExpElement > newRegExp ( regexp.getRegExp().clone() );

	for(const auto& symbol : string.getContent()) {
		std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > > out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpDerivation::REGEXP_DERIVATION);
		newRegExp = std::move ( out.second );
	}

	return regexp::FormalRegExp ( std::move ( * newRegExp ) );
}

auto RegExpDerivationFormalRegExp = RegExpDerivation::RegistratorWrapper<regexp::FormalRegExp, regexp::FormalRegExp>(RegExpDerivation::derivation);

regexp::UnboundedRegExp RegExpDerivation::derivation(const regexp::UnboundedRegExp& regexp, const string::LinearString& string) {
	std::unique_ptr < regexp::UnboundedRegExpElement > newRegExp ( regexp.getRegExp().clone() );

	for(const auto& symbol : string.getContent()) {
		std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpDerivation::REGEXP_DERIVATION);
		newRegExp = std::move ( out.second );
	}

	return regexp::UnboundedRegExp (std::move (* newRegExp));
}

auto RegExpDerivationUnboundedRegExp = RegExpDerivation::RegistratorWrapper<regexp::UnboundedRegExp, regexp::UnboundedRegExp>(RegExpDerivation::derivation);

// ----------------------------------------------------------------------------

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpAlternation& alternation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	alternation.getLeftElement().Accept(userData, *this);
	std::unique_ptr < regexp::FormalRegExpElement > leftDerivative = std::move ( out.second );

	alternation.getRightElement().Accept(userData, *this);

	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpAlternation( std::move ( * leftDerivative ), std::move ( * out.second ) ) );
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpConcatenation& concatenation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	concatenation.getLeftElement().Accept(userData, *this);

	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpConcatenation( std::move ( * out.second ), std::move_copy ( concatenation.getRightElement ( ) ) ) );

	if(regexp::properties::RegExpEpsilon::languageContainsEpsilon(concatenation.getLeftElement())) {
		std::unique_ptr < regexp::FormalRegExpElement > alternation = std::move ( out.second );
		concatenation.getRightElement().Accept(userData, *this);

		out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpAlternation ( std::move ( * alternation ), std::move( * out.second ) ) );
	}
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpIteration& iteration) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	iteration.getElement().Accept(userData, *this);

	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpConcatenation( std::move ( * out.second ), iteration ) );
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpSymbol& symbol) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	if(out.first == symbol.getSymbol())
		out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpEpsilon() );
	else
		out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpEmpty() );
}


void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpEpsilon&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);
	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpEmpty() );
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpEmpty&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);
	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpEmpty() );
}

// ----------------------------------------------------------------------------


void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpAlternation& alternation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);
	std::unique_ptr < regexp::UnboundedRegExpAlternation > ret ( new regexp::UnboundedRegExpAlternation() );

	for(const auto& child : alternation.getElements()) {
		child->Accept(userData, *this);
		ret->appendElement(std::move( * out.second ) );
	}

	out.second = std::move ( ret );
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpConcatenation& concatenation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);
	std::unique_ptr < regexp::UnboundedRegExpAlternation > ret ( new regexp::UnboundedRegExpAlternation() );

	for(auto child = concatenation.getElements().begin(); child != concatenation.getElements().end(); ++ child) {
		std::unique_ptr < regexp::UnboundedRegExpConcatenation > concat ( new regexp::UnboundedRegExpConcatenation() );
		(*child)->Accept(userData, *this);
		concat->appendElement( std::move ( * out.second ) );

		auto succeedingElement = child;
		while(++succeedingElement != concatenation.getElements().end())
			concat->appendElement(**succeedingElement);

		ret->appendElement(std::move(*concat));

		if( ! regexp::properties::RegExpEpsilon::languageContainsEpsilon(**child) )
			break;
	}

	out.second = std::move ( ret );
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpIteration& iteration) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);

	iteration.getElement().Accept(userData, *this);

	UnboundedRegExpConcatenation * con = new regexp::UnboundedRegExpConcatenation( );
	con->appendElement ( std::move ( * out.second ) );
	con->appendElement ( iteration );
	out.second = std::unique_ptr < UnboundedRegExpElement > ( con );
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpSymbol& symbol) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);

	if(out.first == symbol.getSymbol())
		out.second = std::unique_ptr < UnboundedRegExpElement > ( new regexp::UnboundedRegExpEpsilon() );
	else
		out.second = std::unique_ptr < UnboundedRegExpElement > ( new regexp::UnboundedRegExpEmpty() );
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpEpsilon&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);
	out.second = std::unique_ptr < UnboundedRegExpElement > ( new regexp::UnboundedRegExpEmpty() );
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpEmpty&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>>*) userData);
	out.second = std::unique_ptr < UnboundedRegExpElement > ( new regexp::UnboundedRegExpEmpty() );
}

const RegExpDerivation RegExpDerivation::REGEXP_DERIVATION;

} /* namespace regexp */
