/*
 * RegExpIntegral.cpp
 *
 *  Created on: 24. 2. 2014
 *	  Author: Tomas Pecka
 */

#include "RegExpIntegral.h"
#include <regexp/unbounded/UnboundedRegExpElements.h>
#include <regexp/formal/FormalRegExpElements.h>

namespace regexp{

regexp::RegExp RegExpIntegral::integral(const regexp::RegExp& regexp, const string::LinearString& string) {
	return dispatch(regexp.getData(), string);
}

regexp::FormalRegExp RegExpIntegral::integral(const regexp::FormalRegExp& regexp, const string::LinearString& string) {
	std::unique_ptr < regexp::FormalRegExpElement > newRegExp ( regexp.getRegExp().clone() );

	for(const auto& symbol : string.getContent()) {
		std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > > out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpIntegral::REGEXP_INTEGRAL);
		newRegExp = std::move ( out.second );
	}

	return regexp::FormalRegExp (* newRegExp);
}

auto RegExpIntegralFormalRegExp = RegExpIntegral::RegistratorWrapper<regexp::FormalRegExp, regexp::FormalRegExp>(RegExpIntegral::integral);

regexp::UnboundedRegExp RegExpIntegral::integral(const regexp::UnboundedRegExp& regexp, const string::LinearString& string) {
	std::unique_ptr < regexp::UnboundedRegExpElement > newRegExp ( regexp.getRegExp().clone() );

	for(const auto& symbol : string.getContent()) {
		std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > > out(symbol, nullptr);
		newRegExp->Accept((void*) &out, RegExpIntegral::REGEXP_INTEGRAL);
		newRegExp = std::move ( out.second );
	}

	return regexp::UnboundedRegExp (* newRegExp);
}

auto RegExpIntegralUnboundedRegExp = RegExpIntegral::RegistratorWrapper<regexp::UnboundedRegExp, regexp::UnboundedRegExp>(RegExpIntegral::integral);

// ----------------------------------------------------------------------------

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpAlternation& alternation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	alternation.getLeftElement().Accept(userData, *this);
	std::unique_ptr < regexp::FormalRegExpElement > leftIntegral = std::move ( out.second );

	alternation.getRightElement().Accept(userData, *this);

	out.second = std::unique_ptr < FormalRegExpElement > ( new regexp::FormalRegExpAlternation( std::move ( * leftIntegral ), std::move ( * out.second ) ) );
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpConcatenation& concatenation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	out.second = std::unique_ptr < regexp::FormalRegExpElement > ( new regexp::FormalRegExpConcatenation ( regexp::FormalRegExpSymbol ( out.first ), concatenation ) );
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpIteration& iteration) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	out.second = std::unique_ptr < regexp::FormalRegExpElement > ( new regexp::FormalRegExpConcatenation ( regexp::FormalRegExpSymbol ( out.first ), iteration ) );
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpSymbol& symbol) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);

	out.second = std::unique_ptr < regexp::FormalRegExpElement > ( new regexp::FormalRegExpConcatenation ( regexp::FormalRegExpSymbol ( out.first ), symbol ) );
}


void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpEpsilon&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);
	out.second = std::unique_ptr < regexp::FormalRegExpElement > ( new regexp::FormalRegExpSymbol( out.first ) );
}

void RegExpIntegral::Visit(void* userData, const regexp::FormalRegExpEmpty&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::FormalRegExpElement > >*) userData);
	out.second = std::unique_ptr < regexp::FormalRegExpElement > ( new regexp::FormalRegExpEmpty( ) );
}

// ----------------------------------------------------------------------------

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpAlternation& alternation) const
{
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	regexp::UnboundedRegExpAlternation* alt = new regexp::UnboundedRegExpAlternation();

	for(const auto& child : alternation.getElements()) {
		child->Accept(userData, *this);
		alt->appendElement(std::move(*out.second));
	}

	out.second = std::unique_ptr < UnboundedRegExpElement > ( alt );
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpConcatenation& concatenation) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	regexp::UnboundedRegExpConcatenation * con = new regexp::UnboundedRegExpConcatenation( );

	con->appendElement ( regexp::UnboundedRegExpSymbol ( out.first ) );
	for ( const auto & element : concatenation.getElements() )
		con->appendElement ( * element );

	out.second = std::unique_ptr < regexp::UnboundedRegExpElement > ( con );
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpIteration& iteration) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	regexp::UnboundedRegExpConcatenation * con = new regexp::UnboundedRegExpConcatenation( );
	con->appendElement ( regexp::UnboundedRegExpSymbol ( out.first ) );
	con->appendElement ( iteration );

	out.second = std::unique_ptr < regexp::UnboundedRegExpElement > ( con );
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpSymbol& symbol) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	regexp::UnboundedRegExpConcatenation * con = new regexp::UnboundedRegExpConcatenation( );
	con->appendElement ( regexp::UnboundedRegExpSymbol ( out.first ) );
	con->appendElement ( symbol );

	out.second = std::unique_ptr < regexp::UnboundedRegExpElement > ( con );
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpEpsilon&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	out.second = std::unique_ptr < regexp::UnboundedRegExpElement > ( new regexp::UnboundedRegExpSymbol( out.first ) );
}

void RegExpIntegral::Visit(void* userData, const regexp::UnboundedRegExpEmpty&) const {
	std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement>> &out = *((std::pair<alphabet::Symbol, std::unique_ptr < regexp::UnboundedRegExpElement > >*) userData);

	out.second = std::unique_ptr < regexp::UnboundedRegExpElement > ( new regexp::UnboundedRegExpEmpty( ) );
}

const RegExpIntegral RegExpIntegral::REGEXP_INTEGRAL;

} /* namespace regexp */
