#include <list>
#include "determinizeTest.h"

#include <automaton/FSM/NFA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <alphabet/BottomOfTheStackSymbol.h>

#include "automaton/determinize/Determinize.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( determinizeTest, "automaton" );
CPPUNIT_TEST_SUITE_REGISTRATION( determinizeTest );

void determinizeTest::setUp() {
}

void determinizeTest::tearDown() {
}

void determinizeTest::testDeterminizeNFA() {
  automaton::NFA automaton(automaton::State(1));

  automaton.addState(automaton::State(1));
  automaton.addState(automaton::State(2));
  automaton.addState(automaton::State(3));
  automaton.addInputSymbol(alphabet::symbolFrom("a"));
  automaton.addInputSymbol(alphabet::symbolFrom("b"));

  automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2));
  automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), automaton::State(1));

  automaton.addFinalState(automaton::State(3));

  automaton::DFA determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 3);

}

void determinizeTest::testDeterminizeIDPDA() {
  automaton::InputDrivenNPDA automaton(automaton::State(1), alphabet::symbolFrom('S'));

  automaton.addInputSymbol(alphabet::symbolFrom("a"));
  automaton.addInputSymbol(alphabet::symbolFrom("b"));

  automaton.setPushdownStoreOperation(alphabet::symbolFrom("a"), std::vector<alphabet::Symbol>{}, std::vector<alphabet::Symbol> {});
  automaton.setPushdownStoreOperation(alphabet::symbolFrom("b"), std::vector<alphabet::Symbol>{}, std::vector<alphabet::Symbol> {});

  automaton.addState(automaton::State(1));
  automaton.addState(automaton::State(2));
  automaton.addState(automaton::State(3));
  automaton.addTransition(automaton::State(1), alphabet::symbolFrom("a"), automaton::State(2));
  automaton.addTransition(automaton::State(2), alphabet::symbolFrom("b"), automaton::State(1));

  automaton.addFinalState(automaton::State(3));

  automaton::InputDrivenDPDA determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 3);
}

void determinizeTest::testDeterminizeVPA() {
  automaton::VisiblyPushdownNPDA automaton(alphabet::Symbol(alphabet::BottomOfTheStackSymbol {}));

  automaton.addCallInputSymbol(alphabet::symbolFrom('a'));
  automaton.addReturnInputSymbol(alphabet::symbolFrom('^'));

  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('A'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('B'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('C'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('D'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('E'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('F'));
  automaton.addPushdownStoreSymbol(alphabet::symbolFrom('T'));

  automaton.addState(automaton::State(0));
  automaton.addState(automaton::State(1));
  automaton.addState(automaton::State(2));
  automaton.addState(automaton::State(3));
  automaton.addState(automaton::State(4));
  automaton.addState(automaton::State(5));
  automaton.addState(automaton::State(6));


  automaton.addCallTransition(automaton::State(0), alphabet::symbolFrom('a'), automaton::State(0), alphabet::symbolFrom('A'));
  automaton.addReturnTransition(automaton::State(0), alphabet::symbolFrom('^'), alphabet::symbolFrom('A'), automaton::State(0));

  automaton.addCallTransition(automaton::State(0), alphabet::symbolFrom('a'), automaton::State(1), alphabet::symbolFrom('B'));

  automaton.addCallTransition(automaton::State(1), alphabet::symbolFrom('a'), automaton::State(2), alphabet::symbolFrom('C'));
  automaton.addCallTransition(automaton::State(2), alphabet::symbolFrom('a'), automaton::State(2), alphabet::symbolFrom('D'));
  automaton.addReturnTransition(automaton::State(2), alphabet::symbolFrom('^'), alphabet::symbolFrom('D'), automaton::State(2));
  automaton.addReturnTransition(automaton::State(2), alphabet::symbolFrom('^'), alphabet::symbolFrom('C'), automaton::State(3));

  automaton.addCallTransition(automaton::State(3), alphabet::symbolFrom('a'), automaton::State(4), alphabet::symbolFrom('E'));
  automaton.addCallTransition(automaton::State(4), alphabet::symbolFrom('a'), automaton::State(4), alphabet::symbolFrom('F'));
  automaton.addReturnTransition(automaton::State(4), alphabet::symbolFrom('^'), alphabet::symbolFrom('F'), automaton::State(4));
  automaton.addReturnTransition(automaton::State(4), alphabet::symbolFrom('^'), alphabet::symbolFrom('E'), automaton::State(5));

  automaton.addReturnTransition(automaton::State(5), alphabet::symbolFrom('^'), alphabet::symbolFrom('B'), automaton::State(6));

  automaton.addInitialState(automaton::State(0));
  automaton.addFinalState(automaton::State(4));
}

void determinizeTest::testDeterminizeNFTA() {
  automaton::NFTA automaton;

  const alphabet::RankedSymbol a ('a', 2);
  const alphabet::RankedSymbol b ('b', 1);
  const alphabet::RankedSymbol c ('c', 0);
  const std::set<alphabet::RankedSymbol> alphabet {a, b, c};
  automaton.setInputAlphabet(alphabet);

  automaton.addState(automaton::State(1));
  automaton.addState(automaton::State(2));
  automaton.addState(automaton::State(3));

  std::vector<automaton::State> a1States = {automaton::State(1), automaton::State(3)};
  automaton.addTransition(a, a1States, automaton::State(1));
  automaton.addTransition(a, a1States, automaton::State(3));
  std::vector<automaton::State> a2States = {automaton::State(3), automaton::State(3)};
  automaton.addTransition(a, a2States, automaton::State(2));
  std::vector<automaton::State> bStates = {automaton::State(2)};
  automaton.addTransition(b, bStates, automaton::State(1));
  std::vector<automaton::State> cStates;
  automaton.addTransition(c, cStates, automaton::State(3));

  automaton.addFinalState(automaton::State(3));


  automaton::DFTA determinized = automaton::determinize::Determinize::determinize(automaton);

  CPPUNIT_ASSERT(determinized.getStates().size() == 5);
  CPPUNIT_ASSERT(determinized.getFinalStates().size() == 3);
  CPPUNIT_ASSERT(determinized.getTransitions().size() == 15);
}
