#include "LR0Parser.h"

#include "grammar/parsing/LRParser.h"
#include "grammar/parsing/LR0Parser.h"

#include <label/LR0ItemsLabel.h>

#include <vector>

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION ( LR0Parser, "grammar" );
CPPUNIT_TEST_SUITE_REGISTRATION ( LR0Parser );

void LR0Parser::setUp ( ) {
}

void LR0Parser::tearDown ( ) {
}

grammar::CFG LR0Parser::getExpressionGrammar() {
	grammar::CFG expressionGrammar(
		{ E, T, F },
		{ plus, times, leftParenthesis, rightParenthesis, identifier },
		E
	);
	
	expressionGrammar.addRule ( { E }, { E, plus, T } );
	expressionGrammar.addRule ( { E }, { T } );
	expressionGrammar.addRule ( { T }, { T, times, F } );
	expressionGrammar.addRule ( { T }, { F } );
	expressionGrammar.addRule ( { F }, { leftParenthesis, E, rightParenthesis } );
	expressionGrammar.addRule ( { F }, { identifier } );
	
	return expressionGrammar;
}

void LR0Parser::testClosure ( ) {
	grammar::parsing::LR0Items kernelItems {
		{
			E,
			{
				{ 2 , { E, plus, T } }
			}
		},
	};

	grammar::parsing::LR0Items closure = grammar::parsing::LR0Parser::getClosure( kernelItems, getExpressionGrammar ( ) );
	
	grammar::parsing::LR0Items correctClosure {
		{
			E,
			{
				{ 2 , { E, plus, T } }
			}
		},
		{
			T,
			{
				{ 0 , { T, times, F } },
				{ 0 , { F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};

	CPPUNIT_ASSERT ( closure == correctClosure );
}

void LR0Parser::testNextStateItems ( ) {
	grammar::parsing::LR0Items currentStateItems {
		{
			E,
			{
				{ 1 , { E, plus, T } }
			}
		},
	};
	
	grammar::parsing::LR0Items nextStateItems = grammar::parsing::LR0Parser::getNextStateItems( currentStateItems, plus, getExpressionGrammar ( ) );
	
	grammar::parsing::LR0Items correctNextStateItems {
		{
			E,
			{
				{ 2 , { E, plus, T } }
			}
		},
		{
			T,
			{
				{ 0 , { T, times, F } },
				{ 0 , { F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};
	
	CPPUNIT_ASSERT ( nextStateItems == correctNextStateItems );
}

void LR0Parser::testAutomaton ( ) {
	std::vector < grammar::parsing::LR0Items > items(12);
	
	grammar::CFG augmentedExpressionGrammar = grammar::parsing::LRParser::getAugmentedGrammar ( getExpressionGrammar ( ) );
	
	items[0] = {
		{
			augmentedExpressionGrammar.getInitialSymbol(),
			{
				{ 0 , { E } }
			}
		},
		{
			E,
			{
				{ 0 , { E, plus, T } },
				{ 0 , { T } }
			}
		},
		{
			T,
			{
				{ 0 , { T, times, F } },
				{ 0 , { F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};
	
	items[1] = {
		{
			augmentedExpressionGrammar.getInitialSymbol(),
			{
				{ 1 , { E } }
			}
		},
		{
			E,
			{
				{ 1 , { E, plus, T } }
			}
		}
	};

	items[2] = {
		{
			E,
			{
				{ 1 , { T } }
			}
		},
		{
			T,
			{
				{ 1 , { T, times, F } },
			}
		}
	};

	items[3] = {
		{
			T,
			{
				{ 1 , { F } }
			}
		}
	};

	items[4] = {
		{
			E,
			{
				{ 0 , { E, plus, T } },
				{ 0 , { T } }
			}
		},
		{
			T,
			{
				{ 0 , { T, times, F } },
				{ 0 , { F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 1 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};

	items[5] = {
		{
			F,
			{
				{ 1 , { identifier } }
			}
		}
	};
	
	items[6] = {
		{
			E,
			{
				{ 2 , { E, plus, T } },
			}
		},
		{
			T,
			{
				{ 0 , { T, times, F } },
				{ 0 , { F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};
	
	items[7] = {
		{
			T,
			{
				{ 2 , { T, times, F } }
			}
		},
		{
			F,
			{
				{ 0 , { leftParenthesis, E, rightParenthesis } },
				{ 0 , { identifier } }
			}
		}
	};
	
	items[8] = {
		{
			E,
			{
				{ 1 , { E, plus, T } }
			}
		},
		{
			F,
			{
				{ 2 , { leftParenthesis, E, rightParenthesis } },
			}
		}
	};
	
	items[9] = {
		{
			E,
			{
				{ 3 , { E, plus, T } }
			}
		},
		{
			T,
			{
				{ 1 , { T, times, F } }
			}
		}
	};
	
	items[10] = {
		{
			T,
			{
				{ 3 , { T, times, F } }
			}
		}
	};
	
	items[11] = {
		{
			F,
			{
				{ 3 , { leftParenthesis, E, rightParenthesis } }
			}
		}
	};
	
	std::vector < automaton::State > states;
	for (const grammar::parsing::LR0Items & currentItems : items) {
		states.emplace_back( label::Label ( label::LR0ItemsLabel ( currentItems ) ) );
	}
	
	automaton::DFA parsingAutomaton ( states[0] );
	parsingAutomaton.addInputSymbols ( augmentedExpressionGrammar.getNonterminalAlphabet ( ) );
	parsingAutomaton.addInputSymbols ( augmentedExpressionGrammar.getTerminalAlphabet ( ) );
	
	for (const automaton::State & state : states) {
		parsingAutomaton.addState(state);
	}
	
	parsingAutomaton.addTransition(states[0], E, states[1]);
	parsingAutomaton.addTransition(states[0], T, states[2]);
	parsingAutomaton.addTransition(states[0], identifier, states[5]);
	parsingAutomaton.addTransition(states[0], leftParenthesis, states[4]);
	parsingAutomaton.addTransition(states[0], F, states[3]);

	parsingAutomaton.addTransition(states[1], plus, states[6]);
	
	parsingAutomaton.addTransition(states[2], times, states[7]);
	
	parsingAutomaton.addTransition(states[4], E, states[8]);
	parsingAutomaton.addTransition(states[4], F, states[3]);
	parsingAutomaton.addTransition(states[4], T, states[2]);
	parsingAutomaton.addTransition(states[4], leftParenthesis, states[4]);
	parsingAutomaton.addTransition(states[4], identifier, states[5]);
	
	parsingAutomaton.addTransition(states[6], T, states[9]);
	parsingAutomaton.addTransition(states[6], F, states[3]);
	parsingAutomaton.addTransition(states[6], leftParenthesis, states[4]);
	parsingAutomaton.addTransition(states[6], identifier, states[5]);

	parsingAutomaton.addTransition(states[7], F, states[10]);
	parsingAutomaton.addTransition(states[7], leftParenthesis, states[4]);
	parsingAutomaton.addTransition(states[7], identifier, states[5]);

	parsingAutomaton.addTransition(states[8], rightParenthesis, states[11]);
	parsingAutomaton.addTransition(states[8], plus, states[6]);

	parsingAutomaton.addTransition(states[9], times, states[7]);
	
	CPPUNIT_ASSERT ( grammar::parsing::LR0Parser::getAutomaton(getExpressionGrammar()) == parsingAutomaton );
}
