/*
 * CompactNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef COMPACT_DFA_H_
#define COMPACT_DFA_H_

#include "../AutomatonException.h"
#include <map>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/Symbol.h"
#include "../../string/LinearString.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

class CompactNFA : public AutomatonBase, public std::Components < CompactNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < State, string::LinearString >, std::set < State > > transitions;

public:
	explicit CompactNFA ( State initialState );
	explicit CompactNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates );
	explicit CompactNFA ( const EpsilonNFA & other );
	explicit CompactNFA ( const MultiInitialStateNFA & other );
	explicit CompactNFA ( const NFA & other );
	explicit CompactNFA ( const DFA & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const State & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( State state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( State current, string::LinearString input, State next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const State & current, const string::LinearString & input, const State & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < State, string::LinearString >, std::set < State > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::pair < State, string::LinearString >, std::set < State > > getTransitionsFromState ( const State & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < State, string::LinearString >, std::set < State > > getTransitionsToState ( const State & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CompactNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "CompactNFA";

		return xmlTagName;
	}

	static CompactNFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, CompactNFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::CompactNFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::CompactNFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < automaton::State, ::string::LinearString >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second.getAlphabet ( ).count ( symbol ) )
				return true;

		return false;
	}

	static bool available ( const automaton::CompactNFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::CompactNFA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < automaton::State, ::string::LinearString >, std::set < automaton::State > > & t : automaton.getTransitions ( ) )
			if ( ( t.first.first == state ) || t.second.count ( state ) )
				return true;

		return false;
	}

	static bool available ( const automaton::CompactNFA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::CompactNFA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::CompactNFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint< automaton::CompactNFA, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::CompactNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::CompactNFA &, const automaton::State & ) {
	}
};

} /* namespace std */

#endif /* COMPACT_DFA_H_ */
