/*
 * MultiInitialStateNFA.cpp
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace automaton {

MultiInitialStateNFA::MultiInitialStateNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, std::set < State > initialStates, std::set < State > finalStates ) : std::Components < MultiInitialStateNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

MultiInitialStateNFA::MultiInitialStateNFA ( ) : MultiInitialStateNFA ( std::set < State > { }, std::set < alphabet::Symbol > { }, std::set < State > { }, std::set < State > { } ) {
}

MultiInitialStateNFA::MultiInitialStateNFA ( const DFA & other ) : MultiInitialStateNFA ( other.getStates ( ), other.getInputAlphabet ( ), { other.getInitialState ( ) }, other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions[transition.first].insert ( transition.second );
}

MultiInitialStateNFA::MultiInitialStateNFA ( const NFA & other ) : MultiInitialStateNFA ( other.getStates ( ), other.getInputAlphabet ( ), { other.getInitialState ( ) }, other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions[transition.first] = transition.second;
}

AutomatonBase * MultiInitialStateNFA::clone ( ) const {
	return new MultiInitialStateNFA ( * this );
}

AutomatonBase * MultiInitialStateNFA::plunder ( ) && {
	return new MultiInitialStateNFA ( std::move ( * this ) );
}

bool MultiInitialStateNFA::addTransition ( State from, alphabet::Symbol input, State to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + ( std::string ) input + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist." );

	std::pair < State, alphabet::Symbol > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

bool MultiInitialStateNFA::removeTransition ( const State & from, const alphabet::Symbol & input, const State & to ) {
	std::pair < State, alphabet::Symbol > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

const std::map < std::pair < State, alphabet::Symbol >, std::set < State > > & MultiInitialStateNFA::getTransitions ( ) const {
	return transitions;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > MultiInitialStateNFA::getTransitionsFromState ( const State & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ( std::string ) from.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsFromState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

std::map < std::pair < State, alphabet::Symbol >, std::set < State > > MultiInitialStateNFA::getTransitionsToState ( const State & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ( std::string ) to.getName ( ) + "\" doesn't exist" );

	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitionsToState;

	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

bool MultiInitialStateNFA::isDeterministic ( ) const {
	for ( const std::pair < const std::pair < State, alphabet::Symbol >, std::set < State > > & transition : transitions )
		if ( transition.second.size ( ) > 1 )
			return false;

	return getInitialStates ( ).size ( ) == 1;
}

bool MultiInitialStateNFA::isTotal ( ) const {
	return isDeterministic ( ) && transitionsSize ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

unsigned MultiInitialStateNFA::transitionsSize ( ) const {
	int res = 0;

	for ( const auto & transition : transitions )
		res += transition.second.size ( );

	return res;
}

int MultiInitialStateNFA::compare ( const MultiInitialStateNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialStates ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialStates ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void MultiInitialStateNFA::operator >>( std::ostream & out ) const {
	out << "(MultiInitialStateNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialStates = " << getInitialStates ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

MultiInitialStateNFA::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

MultiInitialStateNFA MultiInitialStateNFA::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, MultiInitialStateNFA::getXmlTagName() );

	std::set < State > states = AutomatonFromXMLParser::parseStates ( input );
	std::set < alphabet::Symbol > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet ( input );
	std::set < State > initialStates = AutomatonFromXMLParser::parseInitialStates ( input );
	std::set < State > finalStates = AutomatonFromXMLParser::parseFinalStates ( input );

	MultiInitialStateNFA automaton;

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setInitialStates ( std::move ( initialStates ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < MultiInitialStateNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, MultiInitialStateNFA::getXmlTagName() );
	return automaton;
}

void MultiInitialStateNFA::parseTransition ( std::deque < sax::Token >::iterator & input, MultiInitialStateNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	State from = AutomatonFromXMLParser::parseTransitionFrom ( input );
	alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol ( input );
	State to = AutomatonFromXMLParser::parseTransitionTo ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputSymbol ), std::move ( to ) );
}

void MultiInitialStateNFA::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( MultiInitialStateNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialStates ( out, this->getInitialStates ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( MultiInitialStateNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void MultiInitialStateNFA::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputSymbol ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace automaton */

namespace alib {

auto multiInitialStateNFAParserRegister = xmlApi < automaton::Automaton >::ParserRegister < automaton::MultiInitialStateNFA > ( );
auto multiInitialStateNFAParserRegister2 = xmlApi < alib::Object >::ParserRegister < automaton::MultiInitialStateNFA > ( );

auto MultiInitialStateNFAFromDFA = castApi::CastRegister < automaton::MultiInitialStateNFA, automaton::DFA > ( );
auto MultiInitialStateNFAFromNFA = castApi::CastRegister < automaton::MultiInitialStateNFA, automaton::NFA > ( );
auto MultiInitialStateNFACastBinder = castApi::CastPoolStringBinder < automaton::MultiInitialStateNFA > ( automaton::MultiInitialStateNFA::getXmlTagName() );

} /* namespace alib */
