/*
 * MultiInitialStateNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef MULTI_INITIAL_STATE_NFA_H_
#define MULTI_INITIAL_STATE_NFA_H_

#include "../AutomatonException.h"
#include <map>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/Symbol.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialStates;

class MultiInitialStateNFA : public AutomatonBase, public std::Components < MultiInitialStateNFA, alphabet::Symbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > transitions;

public:
	explicit MultiInitialStateNFA ( );
	explicit MultiInitialStateNFA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, std::set < State > initialStates, std::set < State > finalStates );
	explicit MultiInitialStateNFA ( const NFA & other );
	explicit MultiInitialStateNFA ( const DFA & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < State > & getInitialStates ( ) const {
		return accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( State state ) {
		return accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < State > states ) {
		accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( State current, alphabet::Symbol input, State next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const State & current, const alphabet::Symbol & input, const State & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < State, alphabet::Symbol >, std::set < State > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > getTransitionsFromState ( const State & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < State, alphabet::Symbol >, std::set < State > > getTransitionsToState ( const State & from ) const;

	/**
	 * Determines whether MultiInitialStateNFA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c contains only 1 initial state.
	 * \li \c is epsilon free. Trivial for this class
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	/**
	 * Determines whether MultiInitialStateNFA is total deterministic
	 * FA is deterministic if and only if
	 * \li \c it is deterministic
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 * @return true when automaton is total deterministic, false otherwise
	 */
	bool isTotal ( ) const;

	unsigned transitionsSize ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const MultiInitialStateNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "MultiInitialStateNFA";

		return xmlTagName;
	}

	static MultiInitialStateNFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, MultiInitialStateNFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::MultiInitialStateNFA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::MultiInitialStateNFA & automaton, const alphabet::Symbol & symbol ) {
		for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second == symbol )
				return true;

		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::MultiInitialStateNFA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::MultiInitialStateNFA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::MultiInitialStateNFA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < automaton::State, alphabet::Symbol >, std::set < automaton::State > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
				return true;

		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::MultiInitialStateNFA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::MultiInitialStateNFA, automaton::State, automaton::InitialStates > {
public:
	static bool used ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::MultiInitialStateNFA &, const automaton::State & ) {
	}
};

} /* namespace std */

#endif /* MULTI_INITIAL_STATE_NFA_H_ */
