/*
 * DPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef DPDA_H_
#define DPDA_H_

#include "../AutomatonException.h"
#include <set>
#include <map>
#include <vector>
#include <variant>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, \alpha) \neq \emptyset$, $\delta (q, a, \beta) \neq \emptyset$ and $\alpha \neq \beta$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (formally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 * if $\delta(q, a, \alpha) \neq \emptyset$, $\delta (q, \varepsilon, \beta) \neq \emptyset$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (fornally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 */
class DPDA : public AutomatonBase, public std::Components < DPDA, alphabet::Symbol, std::tuple < InputAlphabet, PushdownStoreAlphabet >, std::tuple < InitialSymbol >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::pair < State, std::vector < alphabet::Symbol > > > transitions;

public:
	explicit DPDA ( std::set < State > states, std::set < alphabet::Symbol > inputAlphabet, std::set < alphabet::Symbol > pushdownStoreSymbol, State initialState, alphabet::Symbol initialPushdownSymbol, std::set < automaton::State > finalStates );
	explicit DPDA ( State initialState, alphabet::Symbol initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const State & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( State state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getPushdownStoreAlphabet ( ) const {
		return accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getInitialSymbol ( ) const {
		return accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( alphabet::Symbol symbol ) {
		return accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the DPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the DPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( State from, std::variant < string::Epsilon, alphabet::Symbol > input, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	bool addTransition ( State from, alphabet::Symbol input, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	bool addTransition ( State from, std::vector < alphabet::Symbol > pop, State to, std::vector < alphabet::Symbol > push );

	/**
	 * Removes the transition from the DPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the DPDA
	 */
	bool removeTransition ( const State & from, const std::variant < string::Epsilon, alphabet::Symbol > & input, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	bool removeTransition ( const State & from, const alphabet::Symbol & input, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	bool removeTransition ( const State & from, const std::vector < alphabet::Symbol > & pop, const State & to, const std::vector < alphabet::Symbol > & push );

	/**
	 * @return DPDA transitions
	 */
	const std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::pair < State, std::vector < alphabet::Symbol > > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::pair < State, std::vector < alphabet::Symbol > > > getTransitionsFromState ( const State & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::tuple < State, std::variant < string::Epsilon, alphabet::Symbol >, std::vector < alphabet::Symbol > >, std::pair < State, std::vector < alphabet::Symbol > > > getTransitionsToState ( const State & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DPDA";

		return xmlTagName;
	}

	static DPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::DPDA, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DPDA & automaton, const alphabet::Symbol & symbol ) {
		for (const std::pair<const std::tuple<automaton::State, std::variant<::string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<automaton::State, std::vector<alphabet::Symbol> > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).is<alphabet::Symbol>() && symbol == std::get<1>(transition.first).get<alphabet::Symbol>())
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::DPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::DPDA, alphabet::Symbol, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::DPDA & automaton, const alphabet::Symbol & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<automaton::State, std::variant<::string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<automaton::State, std::vector<alphabet::Symbol> > >& transition : automaton.getTransitions()) {
			const auto & popSymbols = std::get<2>(transition.first);
			const auto & pushSymbols = transition.second.second;
			if(std::contains(popSymbols.begin(), popSymbols.end(), symbol ) || std::contains(pushSymbols.begin(), pushSymbols.end(), symbol))
				return true;
		}

		return false;
	}

	static bool available ( const automaton::DPDA &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::DPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< automaton::DPDA, alphabet::Symbol, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::DPDA & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::DPDA &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::DPDA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::DPDA & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<automaton::State, std::variant<::string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::pair<automaton::State, std::vector<alphabet::Symbol> > >& transition : automaton.getTransitions())
			if ( state == std::get<0>(transition.first) || transition.second.first == state )
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::DPDA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::DPDA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::DPDA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::DPDA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint< automaton::DPDA, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::DPDA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA &, const automaton::State & ) {
	}
};

} /* namespace std */

#endif /* DPDA_H_ */
