/*
 * VisiblyPushdownDPDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#include "VisiblyPushdownDPDA.h"
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../Automaton.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace automaton {

VisiblyPushdownDPDA::VisiblyPushdownDPDA ( std::set < State > states, std::set < alphabet::Symbol > callAlphabet, std::set < alphabet::Symbol > returnAlphabet, std::set < alphabet::Symbol > localAlphabet, std::set < alphabet::Symbol > pushdownStoreAlphabet, State initialState, alphabet::Symbol bottomOfTheStackSymbol, std::set < State > finalStates ) : std::Components < VisiblyPushdownDPDA, alphabet::Symbol, std::tuple < CallAlphabet, ReturnAlphabet, LocalAlphabet, PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( callAlphabet ), std::move ( returnAlphabet ), std::move ( localAlphabet ), std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

VisiblyPushdownDPDA::VisiblyPushdownDPDA(State initialState, alphabet::Symbol bottomOfTheStackSymbol) : VisiblyPushdownDPDA ( std::set < State > { initialState }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { }, std::set < alphabet::Symbol > { bottomOfTheStackSymbol }, initialState, bottomOfTheStackSymbol, std::set < automaton::State > { }) {
}

AutomatonBase* VisiblyPushdownDPDA::clone() const {
	return new VisiblyPushdownDPDA(*this);
}

AutomatonBase* VisiblyPushdownDPDA::plunder() && {
	return new VisiblyPushdownDPDA(std::move(*this));
}

bool VisiblyPushdownDPDA::addCallTransition(State from, alphabet::Symbol input, State to, alphabet::Symbol push) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (!getCallInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + (std::string) push + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(std::move(to), std::move(push));

	if(callTransitions.find(key) != callTransitions.end() && callTransitions.find(key)->second == value)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	callTransitions.insert(std::make_pair(std::move(key), std::move(value)));
	return true;
}

bool VisiblyPushdownDPDA::addReturnTransition(State from, alphabet::Symbol input, alphabet::Symbol pop, State to) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (!getReturnInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	if (!getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + (std::string) pop + "\" doesn't exist.");
	}

	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(std::move(from), std::move(input), std::move(pop));

	if(returnTransitions.find(key) != returnTransitions.end() && returnTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == std::get<0>(key) && std::get<1>(transition.first) == std::get<1>(key) && std::get<2>(transition.first) == std::get<2>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) from + "\" when transition reading \"" + (std::string) input + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == std::get<0>(key) && transition.first.second == std::get<1>(key))
			throw AutomatonException("Can't add transition from state \"" + (std::string) std::get<0>(key) + "\" when transition reading \"" + (std::string) std::get<1>(key) + "\" is present.");

	returnTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool VisiblyPushdownDPDA::addLocalTransition(State from, alphabet::Symbol input, State to) {
	if (!getStates().count(from)) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (!getLocalInputAlphabet().count(input)) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (!getStates().count(to)) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	std::pair<State, alphabet::Symbol> key(std::move(from), std::move(input));

	if(localTransitions.find(key) != localTransitions.end() && localTransitions.find(key)->second == to)
		return false;

	for(const auto& transition : callTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : returnTransitions)
		if(std::get<0>(transition.first) == key.first && std::get<1>(transition.first) == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	for(const auto& transition : localTransitions)
		if(transition.first.first == key.first && transition.first.second == key.second)
			throw AutomatonException("Can't add transition from state \"" + (std::string) key.first + "\" when transition reading \"" + (std::string) key.second + "\" is present.");

	localTransitions.insert(std::make_pair(std::move(key), std::move(to)));
	return true;
}

bool VisiblyPushdownDPDA::removeCallTransition(const State& from, const alphabet::Symbol& input, const State& to, const alphabet::Symbol& push) {
	std::pair<State, alphabet::Symbol> key(from, input);
	std::pair<automaton::State, alphabet::Symbol> value = std::make_pair(to, push);

	if (callTransitions.find(key) == callTransitions.end())
		return false;

	if(callTransitions.find(key)->second != value)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	callTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeReturnTransition(const State& from, const alphabet::Symbol& input, const alphabet::Symbol& pop, const State& to) {
	std::tuple<State, alphabet::Symbol, alphabet::Symbol> key(from, input, pop);

	if (returnTransitions.find(key) == returnTransitions.end())
		return false;

	if(returnTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	returnTransitions.erase(key);
	return true;
}

bool VisiblyPushdownDPDA::removeLocalTransition(const State& from, const alphabet::Symbol& input, const State& to) {
	std::pair<State, alphabet::Symbol> key(from, input);

	if (localTransitions.find(key) == localTransitions.end())
		return false;

	if(localTransitions.find(key)->second != to)
		throw AutomatonException(
				"Transition (\"" + (std::string) from.getName() + "\", \"" + (std::string) input
						+ "\") -> \"" + (std::string) to.getName() + "\" doesn't exist.");

	localTransitions.erase(key);
	return true;
}

const std::map<std::pair<State, alphabet::Symbol>, std::pair<State, alphabet::Symbol> >& VisiblyPushdownDPDA::getCallTransitions() const {
	return callTransitions;
}

const std::map<std::tuple<State, alphabet::Symbol, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getReturnTransitions() const {
	return returnTransitions;
}

const std::map<std::pair<State, alphabet::Symbol>, State>& VisiblyPushdownDPDA::getLocalTransitions() const {
	return localTransitions;
}

int VisiblyPushdownDPDA::compare(const VisiblyPushdownDPDA& other) const {
	auto first = std::tie(getStates(), getCallInputAlphabet(), getReturnInputAlphabet(), getLocalInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getCallInputAlphabet(), other.getReturnInputAlphabet(), other.getLocalInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	std::compare<decltype(first)> comp;
	return comp(first, second);
}

void VisiblyPushdownDPDA::operator>>(std::ostream& out) const {
	out << "(VisiblyPushdownDPDA "
		<< "states = " << getStates()
		<< "callAlphabet = " << getCallInputAlphabet()
		<< "returnAlphabet = " << getReturnInputAlphabet()
		<< "localAlphabet = " << getLocalInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

VisiblyPushdownDPDA::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

VisiblyPushdownDPDA VisiblyPushdownDPDA::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, VisiblyPushdownDPDA::getXmlTagName());

	std::set<State> states = AutomatonFromXMLParser::parseStates(input);
	std::set<alphabet::Symbol> callAlphabet = AutomatonFromXMLParser::parseCallInputAlphabet(input);
	std::set<alphabet::Symbol> returnAlphabet = AutomatonFromXMLParser::parseReturnInputAlphabet(input);
	std::set<alphabet::Symbol> localAlphabet = AutomatonFromXMLParser::parseLocalInputAlphabet(input);
	std::set<alphabet::Symbol> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet(input);
	State initialState = AutomatonFromXMLParser::parseInitialState(input);
	alphabet::Symbol bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol(input);
	std::set<State> finalStates = AutomatonFromXMLParser::parseFinalStates(input);

	VisiblyPushdownDPDA automaton(std::move(initialState), std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setCallInputAlphabet(std::move(callAlphabet));
	automaton.setReturnInputAlphabet(std::move(returnAlphabet));
	automaton.setLocalInputAlphabet(std::move(localAlphabet));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<VisiblyPushdownDPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, VisiblyPushdownDPDA::getXmlTagName());
	return automaton;
}

void VisiblyPushdownDPDA::parseTransition(std::deque<sax::Token>::iterator& input, VisiblyPushdownDPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);
		alphabet::Symbol push = AutomatonFromXMLParser::parseTransitionSinglePush(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		alphabet::Symbol pop = AutomatonFromXMLParser::parseTransitionSinglePop(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		State from = AutomatonFromXMLParser::parseTransitionFrom(input);
		alphabet::Symbol inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol(input);
		State to = AutomatonFromXMLParser::parseTransitionTo(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

void VisiblyPushdownDPDA::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(VisiblyPushdownDPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeCallInputAlphabet(out, this->getCallInputAlphabet());
	AutomatonToXMLComposer::composeReturnInputAlphabet(out, this->getReturnInputAlphabet());
	AutomatonToXMLComposer::composeLocalInputAlphabet(out, this->getLocalInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(VisiblyPushdownDPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

void VisiblyPushdownDPDA::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionSinglePush(out, transition.second.second);

		out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getReturnTransitions()) {
		out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
	}
	for(const auto& transition : this->getLocalTransitions()) {
		out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

} /* namespace automaton */

namespace alib {

auto visiblyPushdownDPDAParserRegister = xmlApi<automaton::Automaton>::ParserRegister<automaton::VisiblyPushdownDPDA>();
auto visiblyPushdownDPDAParserRegister2 = xmlApi<alib::Object>::ParserRegister<automaton::VisiblyPushdownDPDA>();

} /* namespace alib */
