/*
 * DFTA.h
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#ifndef DFTA_H_
#define DFTA_H_

#include "../AutomatonException.h"
#include <map>
#include <vector>
#include <core/components.hpp>
#include "../AutomatonBase.h"
#include "../common/State.h"
#include "../../alphabet/RankedSymbol.h"

namespace automaton {

class InputAlphabet;
class States;
class FinalStates;

/**
 * Represents Finite Tree Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class DFTA : public AutomatonBase, public std::Components < DFTA, alphabet::RankedSymbol, std::tuple < InputAlphabet >, std::tuple < >, automaton::State, std::tuple < States, FinalStates >, std::tuple < > > {
	std::map < std::pair < alphabet::RankedSymbol, std::vector < State > >, State > transitions;

public:
	explicit DFTA ( );
	explicit DFTA ( std::set < State > states, std::set < alphabet::RankedSymbol > inputAlphabet, std::set < State > finalStates );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::RankedSymbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::RankedSymbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::RankedSymbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current symbol in node
	 * @param children states of children of current node
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( const alphabet::RankedSymbol & current, const std::vector < State > & children, const State & next );

	/**
	 * Removes transition from the automaton.
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const alphabet::RankedSymbol symbol, const std::vector < State > & states, const State & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < alphabet::RankedSymbol, std::vector < State > >, State > & getTransitions ( ) const { return transitions; }

	std::set < State > getTransitionRightSide ( const alphabet::RankedSymbol symbol, const std::vector < State > & states );

	unsigned transitionsSize ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DFTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DFTA";

		return xmlTagName;
	}

	static DFTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DFTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::DFTA, alphabet::RankedSymbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DFTA & automaton, const alphabet::RankedSymbol & symbol ) {
		for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<automaton::State> >, automaton::State>& t : automaton.getTransitions())
			if (t.first.first == symbol)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const automaton::DFTA &, const alphabet::RankedSymbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::DFTA, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::DFTA & automaton, const automaton::State & state ) {
		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<alphabet::RankedSymbol, std::vector<automaton::State> >, automaton::State>& t : automaton.getTransitions())
			if(std::contains(t.first.second.begin(), t.first.second.end(), state ) || t.second == state)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::DFTA &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::DFTA, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::DFTA &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::DFTA & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DFTA &, const automaton::State & ) {
	}
};

} /* namespace std */

#endif /* DFTA_H_ */
