/*
 * OneTapeDTM.h
 *
 *  Created on: Apr 24, 2013
 *      Author: Martin Zak
 */

#ifndef ONE_TAPE_DTM_H_
#define ONE_TAPE_DTM_H_

#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include <core/components.hpp>
#include "../common/State.h"
#include "../../alphabet/Symbol.h"
#include "../common/Shift.h"

#include <map>
#include <set>
#include <tuple>

namespace automaton {

class TapeAlphabet;
class InputAlphabet;
class BlankSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * One tape turing machine
 */
class OneTapeDTM : public AutomatonBase, public std::Components < OneTapeDTM, alphabet::Symbol, std::tuple < TapeAlphabet, InputAlphabet >, std::tuple < BlankSymbol >, automaton::State, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < State, alphabet::Symbol >, std::tuple < State, alphabet::Symbol, Shift > > transitions;

public:
	explicit OneTapeDTM ( std::set < State > states, std::set < alphabet::Symbol > tapeAlphabet, alphabet::Symbol blankSymbol, std::set< alphabet::Symbol > inputAlphabet, State initialState, std::set < State > finalStates );

	explicit OneTapeDTM ( State initialState, alphabet::Symbol blank );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const State & getInitialState ( ) const {
		return accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( State state ) {
		return accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < State > & getStates ( ) const {
		return accessComponent < States > ( ).get ( );
	}

	bool addState ( State state ) {
		return accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < State > states ) {
		accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const State & state ) {
		accessComponent < States > ( ).remove ( state );
	}

	const std::set < State > & getFinalStates ( ) const {
		return accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( State state ) {
		return accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < State > states ) {
		accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const State & state ) {
		accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < alphabet::Symbol > & getInputAlphabet ( ) const {
		return accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::Symbol > & getTapeAlphabet ( ) const {
		return accessComponent < TapeAlphabet > ( ).get ( );
	}

	bool addTapeSymbol ( alphabet::Symbol symbol ) {
		return accessComponent < TapeAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addTapeSymbols ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TapeAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setTapeAlphabet ( std::set < alphabet::Symbol > symbols ) {
		accessComponent < TapeAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeTapeSymbol ( const alphabet::Symbol & symbol ) {
		accessComponent < TapeAlphabet > ( ).remove ( symbol );
	}

	const alphabet::Symbol & getBlankSymbol ( ) const {
		return accessElement < BlankSymbol > ( ).get ( );
	}

	bool setBlankSymbol ( alphabet::Symbol state ) {
		return accessElement < BlankSymbol > ( ).set ( std::move ( state ) );
	}

	/**
	 * Adds transition to the automaton.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the TM (state, tape symbol) or when transition already exists
	 */
	bool addTransition ( State from, alphabet::Symbol input, State to, alphabet::Symbol output, Shift shift );

	/**
	 * Removes the transition from the TM.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the TM
	 */
	bool removeTransition ( const State & from, const alphabet::Symbol & input, const State & to, const alphabet::Symbol & output, const Shift & shift );

	/**
	 * @return TM transitions
	 */
	const std::map < std::pair < State, alphabet::Symbol >, std::tuple < State, alphabet::Symbol, Shift > > & getTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const OneTapeDTM & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "OneTapeDTM";

		return xmlTagName;
	}

	static OneTapeDTM parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, OneTapeDTM & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;
};

} /* namespace automaton */

namespace std {

template < >
class ComponentConstraint< automaton::OneTapeDTM, alphabet::Symbol, automaton::TapeAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if ( automaton.getBlankSymbol ( ) == symbol )
			return true;

		if ( automaton.getInputAlphabet().count(symbol))
			return true;

		for (const std::pair<const std::pair<automaton::State, alphabet::Symbol>, std::tuple<automaton::State, alphabet::Symbol, automaton::Shift> >& transition : automaton.getTransitions())
			if (symbol == transition.first.second || symbol == std::get<1>(transition.second))
				return true;

		return false;

	}

	static bool available ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
	}
};

template < >
class ComponentConstraint< automaton::OneTapeDTM, alphabet::Symbol, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::OneTapeDTM &, const alphabet::Symbol & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::TapeAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if (symbol == automaton.getBlankSymbol())
			throw automaton::AutomatonException("Input symbol \"" + (std::string) symbol + "\" cannot be blank symbol.");
	}
};

template < >
class ElementConstraint< automaton::OneTapeDTM, alphabet::Symbol, automaton::BlankSymbol > {
public:
	static bool available ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		return automaton.accessComponent < automaton::TapeAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::OneTapeDTM & automaton, const alphabet::Symbol & symbol ) {
		if (automaton.getInputAlphabet().count( symbol ))
			throw automaton::AutomatonException("Blank symbol \"" + (std::string) symbol + "\" cannot be in input alphabet.");
	}
};

template < >
class ComponentConstraint< automaton::OneTapeDTM, automaton::State, automaton::States > {
public:
	static bool used ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<automaton::State, alphabet::Symbol>, std::tuple<automaton::State, alphabet::Symbol, automaton::Shift> >& transition : automaton.getTransitions ( ) )
			if ( state == transition.first.first || state == std::get < 0 > ( transition.second ) )
				return true;

		return false;
	}

	static bool available ( const automaton::OneTapeDTM &, const automaton::State & ) {
		return true;
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

template < >
class ComponentConstraint< automaton::OneTapeDTM, automaton::State, automaton::FinalStates > {
public:
	static bool used ( const automaton::OneTapeDTM &, const automaton::State & ) {
		return false;
	}

	static bool available ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

template < >
class ElementConstraint< automaton::OneTapeDTM, automaton::State, automaton::InitialState > {
public:
	static bool available ( const automaton::OneTapeDTM & automaton, const automaton::State & state ) {
		return automaton.accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::OneTapeDTM &, const automaton::State & ) {
	}
};

} /* namespace std */

#endif /* ONE_TAPE_DTM_H_ */
