/*
 * RightLG.cpp
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#include "RightLG.h"
#include <algorithm>
#include <sstream>

#include "../../alphabet/Symbol.h"

#include <sax/FromXMLParserHelper.h>
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../Grammar.h"
#include <object/Object.h>
#include <XmlApi.hpp>

namespace grammar {

RightLG::RightLG ( alphabet::Symbol initialSymbol ) : RightLG ( std::set < alphabet::Symbol > { initialSymbol }, std::set < alphabet::Symbol > ( ), initialSymbol ) {
}

RightLG::RightLG ( std::set < alphabet::Symbol > nonterminalAlphabet, std::set < alphabet::Symbol > terminalAlphabet, alphabet::Symbol initialSymbol ) : std::Components < RightLG, alphabet::Symbol, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ) {
}

GrammarBase * RightLG::clone ( ) const {
	return new RightLG ( * this );
}

GrammarBase * RightLG::plunder ( ) && {
	return new RightLG ( std::move ( * this ) );
}

bool RightLG::addRule ( alphabet::Symbol leftHandSide, std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.is < std::vector < alphabet::Symbol > > ( ) ) {
		for ( const auto & symbol : rightHandSide.get < std::vector < alphabet::Symbol > > ( ) )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ( std::string ) symbol + " is not a terminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	} else {
		const std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > & rhs = rightHandSide.get < std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > ( );

		if ( !getNonterminalAlphabet ( ).count ( rhs.second ) )
			throw GrammarException ( "Symbol " + ( std::string ) rhs.second + " is not a nonterminal symbol" );

		for ( const auto & symbol : rhs.first )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + ( std::string ) symbol + " is not a terminal symbol" );

		return rules[std::move ( leftHandSide )].insert ( std::move ( rightHandSide ) ).second;
	}
}

bool RightLG::addRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

bool RightLG::addRule ( alphabet::Symbol leftHandSide, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

const std::map < alphabet::Symbol, std::set < std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > > > & RightLG::getRules ( ) const {
	return rules;
}

bool RightLG::removeRule ( const alphabet::Symbol & leftHandSide, const std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

bool RightLG::removeRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

bool RightLG::removeRule ( const alphabet::Symbol & leftHandSide, const std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > & rightHandSide ) {
	std::variant < std::vector < alphabet::Symbol >, std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

bool RightLG::addRawRule ( alphabet::Symbol leftHandSide, std::vector < alphabet::Symbol > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[rightHandSide.size ( ) - 1] ) )
		return addRule ( std::move ( leftHandSide ), std::make_pair ( std::vector < alphabet::Symbol > ( std::make_move_iterator ( rightHandSide.begin ( ) ), std::make_move_iterator ( rightHandSide.end ( ) ) - 1 ), std::move ( rightHandSide[rightHandSide.size ( ) - 1] ) ) );
	else
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
}

std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > RightLG::getRawRules ( ) const {
	std::map < alphabet::Symbol, std::set < std::vector < alphabet::Symbol > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.is < std::vector < alphabet::Symbol > > ( ) ) {
				res[rule.first].insert ( rhs.get < std::vector < alphabet::Symbol > > ( ) );
			} else {
				const auto & rhsTuple = rhs.get < std::pair < std::vector < alphabet::Symbol >, alphabet::Symbol > > ( );
				std::vector < alphabet::Symbol > tmp { rhsTuple.first.begin ( ), rhsTuple.first.end ( ) };
				tmp.push_back ( rhsTuple.second );
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	return res;
}

bool RightLG::removeRawRule ( const alphabet::Symbol & leftHandSide, const std::vector < alphabet::Symbol > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return removeRule ( leftHandSide, rightHandSide );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[rightHandSide.size ( ) - 1] ) )
		return removeRule ( leftHandSide, std::make_pair ( std::vector < alphabet::Symbol > ( rightHandSide.begin ( ), rightHandSide.end ( ) - 1 ), rightHandSide[rightHandSide.size ( ) - 1] ) );
	else
		return removeRule ( leftHandSide, rightHandSide );
}

int RightLG::compare ( const RightLG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RightLG::operator >>( std::ostream & out ) const {
	out << "(LeftLG " << "nonterminalAlphabet = "
	    << getNonterminalAlphabet ( ) << "terminalAlphabet = "
	    << getTerminalAlphabet ( ) << "initialSymbol = "
	    << getInitialSymbol ( ) << "rules = "
	    << rules << ")";
}

RightLG::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

RightLG RightLG::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, RightLG::getXmlTagName() );

	std::set < alphabet::Symbol > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet ( input );
	std::set < alphabet::Symbol > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet ( input );
	alphabet::Symbol initialSymbol = GrammarFromXMLParser::parseInitialSymbol ( input );

	RightLG grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, RightLG::getXmlTagName() );
	return grammar;
}

void RightLG::parseRule ( std::deque < sax::Token >::iterator & input, RightLG & grammar ) {
	alphabet::Symbol lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS ( input );
	std::vector < alphabet::Symbol > rhs = GrammarFromXMLParser::parseRuleRHS ( input );

	grammar.addRawRule ( std::move ( lhs ), std::move ( rhs ) );
}

void RightLG::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( RightLG::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );

	out.emplace_back ( RightLG::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

void RightLG::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleRightLGRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

} /* namespace grammar */

namespace alib {

auto RightLGParserRegister = xmlApi < grammar::Grammar >::ParserRegister < grammar::RightLG > ( );
auto RightLGParserRegister2 = xmlApi < alib::Object >::ParserRegister < grammar::RightLG > ( );

} /* namespace alib */
