/*
 * FormalRegExpElement.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef FORMAL_REG_EXP_ELEMENT_H_
#define FORMAL_REG_EXP_ELEMENT_H_

#include <XmlApi.hpp>
#include "../../alphabet/Symbol.h"
#include <set>
#include <tree>

namespace regexp {

class FormalRegExpAlternation;
class FormalRegExpConcatenation;
class FormalRegExpIteration;
class FormalRegExpSymbol;
class FormalRegExpEmpty;
class FormalRegExpEpsilon;

class UnboundedRegExpElement;

class FormalRegExpElementVisitor {
public:
	virtual void Visit ( void *, const FormalRegExpAlternation & ) const = 0;
	virtual void Visit ( void *, const FormalRegExpConcatenation & ) const = 0;
	virtual void Visit ( void *, const FormalRegExpIteration & ) const = 0;
	virtual void Visit ( void *, const FormalRegExpSymbol & ) const = 0;
	virtual void Visit ( void *, const FormalRegExpEmpty & ) const = 0;
	virtual void Visit ( void *, const FormalRegExpEpsilon & ) const = 0;
};

/**
 * Abstract class representing element in the formal regular expression. Can be operator or symbol.
 */
class FormalRegExpElement : public alib::CommonBase < FormalRegExpElement >, public std::BaseNode < FormalRegExpElement > {
public:
	virtual void Accept ( void * userData, const FormalRegExpElementVisitor & visitor ) const = 0;

	/**
	 * Creates copy of the element.
	 *
	 * @return copy of the element
	 */
	virtual std::smart_ptr < UnboundedRegExpElement > asUnbounded ( ) const = 0;

	/**
	 * Traverses the regexp tree looking if particular Symbol is used in the regexp.
	 *
	 * @param symbol to test if used in regexp element
	 * @return true if symbol is used by the element and its successor
	 */
	virtual bool testSymbol ( const alphabet::Symbol & symbol ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @param alphabet All alphabet symbols encountered are added into this set
	 * @return true if symbol is used by the element and its successor
	 */
	virtual void computeMinimalAlphabet ( std::set < alphabet::Symbol > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree and checks whether all symbols in the regexp tree are in the alphabet
	 *
	 * @param alphabet
	 * @return true if symbols in the regexp are in the alphabet, false otherwise
	 */
	virtual bool checkAlphabet ( const std::set < alphabet::Symbol > & alphabet ) const = 0;

	/**
	 * Traverses the regexp tree computing minimal alphabet needed by regexp
	 *
	 * @return the minimal alphabet needed by the regexp
	 */
	std::set < alphabet::Symbol > computeMinimalAlphabet ( ) const;
};

} /* namespace regexp */

namespace alib {

template < >
struct xmlApi < regexp::FormalRegExpElement > {
	static std::rvalue_ref < regexp::FormalRegExpElement > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static void compose ( std::deque < sax::Token > & output, const regexp::FormalRegExpElement & data );
};

} /* namespace alib */

#endif /* FORMAL_REG_EXP_ELEMENT_H_ */
