/*
 * UnboundedRegExp.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef UNBOUNDED_REG_EXP_H_
#define UNBOUNDED_REG_EXP_H_

#include <string>
#include <set>
#include <core/components.hpp>

#include "../RegExpBase.h"
#include "UnboundedRegExpElement.h"
#include "../../alphabet/Symbol.h"

namespace regexp {

class FormalRegExp;
class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
class UnboundedRegExp : public RegExpBase, public std::Components < UnboundedRegExp, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	std::smart_ptr < UnboundedRegExpElement > regExp;

public:
	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	virtual RegExpBase * clone ( ) const;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	virtual RegExpBase * plunder ( ) &&;

	explicit UnboundedRegExp ( );

	explicit UnboundedRegExp ( std::set < alphabet::Symbol > alphabet, UnboundedRegExpElement && regExp );
	explicit UnboundedRegExp ( std::set < alphabet::Symbol > alphabet, const UnboundedRegExpElement & regExp );
	explicit UnboundedRegExp ( UnboundedRegExpElement && regExp );
	explicit UnboundedRegExp ( const UnboundedRegExpElement & regExp );
	explicit UnboundedRegExp ( const FormalRegExp & other );

	/**
	 * @return Root node of the regular expression tree
	 */
	const UnboundedRegExpElement & getRegExp ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	UnboundedRegExpElement & getRegExp ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRegExp ( UnboundedRegExpElement && regExp );
	void setRegExp ( const UnboundedRegExpElement & regExp );

	virtual const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Prints XML representation of the RegExp to the output stream.
	 * @param out output stream to which print the RegExp
	 * @param regexp RegExp to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const UnboundedRegExp & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "UnboundedRegExp";

		return xmlTagName;
	}

	static UnboundedRegExp parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace regexp */

namespace std {

template < >
class ComponentConstraint< regexp::UnboundedRegExp, alphabet::Symbol, regexp::GeneralAlphabet > {
public:
	static bool used ( const regexp::UnboundedRegExp & regexp, const alphabet::Symbol & symbol ) {
		return regexp.getRegExp ( ).testSymbol ( symbol );
	}

	static bool available ( const regexp::UnboundedRegExp &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const regexp::UnboundedRegExp &, const alphabet::Symbol & ) {
	}
};

} /* namespace std */

#endif /* UNBOUNDED_REG_EXP_H_ */
