#ifndef FORMAL_RTE_H_
#define FORMAL_RTE_H_

#include <string>
#include <set>
#include <exception/CommonException.h>
#include "FormalRTEElement.h"
#include "../RTEBase.h"
#include <core/components.hpp>

namespace rte {

class FormalRTEElement;
class GeneralAlphabet;
class ConstantAlphabet;

// class UnboundedRTE;

/**
 * Represents regular tree expression parsed from the XML. Regular tree expression is stored
 * as a tree of RTEElement elements.
 */
class FormalRTE : public RTEBase, public std::Components < FormalRTE, alphabet::RankedSymbol, std::tuple < GeneralAlphabet, ConstantAlphabet >, std::tuple < > > {
protected:
	std::smart_ptr < FormalRTEElement > rte;

public:
	explicit FormalRTE ( );

	 // explicit FormalRTE ( const UnboundedRTE & other );
	explicit FormalRTE ( std::set < alphabet::RankedSymbol > alphabetF, std::set < alphabet::RankedSymbol > alphabetK, const FormalRTEElement & rte );
	explicit FormalRTE ( std::set < alphabet::RankedSymbol > alphabetF, std::set < alphabet::RankedSymbol > alphabetK, FormalRTEElement && rte );
	explicit FormalRTE ( std::pair < std::set < alphabet::RankedSymbol >, std::set < alphabet::RankedSymbol > > alphabets, const FormalRTEElement & rte );
	explicit FormalRTE ( std::pair < std::set < alphabet::RankedSymbol >, std::set < alphabet::RankedSymbol > > alphabets, FormalRTEElement && rte );
	explicit FormalRTE ( const FormalRTEElement & rte );
	explicit FormalRTE ( FormalRTEElement && rte );

	/**
	 * Copy constructor.
	 * @param other RTE to copy
	 */
	FormalRTE ( const FormalRTE & other );
	FormalRTE ( FormalRTE && other ) noexcept;
	FormalRTE & operator =( const FormalRTE & other );
	FormalRTE & operator =( FormalRTE && other ) noexcept;
	~FormalRTE ( ) noexcept;

	/**
	 * @copydoc FormalRTEElement::clone() const
	 */
	virtual RTEBase * clone ( ) const;

	/**
	 * @copydoc FormalRTEElement::plunder() const
	 */
	virtual RTEBase * plunder ( ) &&;

	const std::set < alphabet::RankedSymbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void addAlphabetSymbols ( alphabet::RankedSymbol symbol ) {
		accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addAlphabetSymbols ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < GeneralAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setAlphabetSymbols ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < GeneralAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeAlphabetSymbol ( const alphabet::RankedSymbol & symbol ) {
		accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	const std::set < alphabet::RankedSymbol > & getConstantAlphabet ( ) const {
		return accessComponent < ConstantAlphabet > ( ).get ( );
	}

	void addConstantSymbol ( alphabet::RankedSymbol symbol ) {
		accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addConstantSymbols ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < ConstantAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setConstantSymbols ( std::set < alphabet::RankedSymbol > symbols ) {
		accessComponent < ConstantAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeConstantSymbol ( const alphabet::RankedSymbol & symbol ) {
		accessComponent < ConstantAlphabet > ( ).remove ( symbol );
	}

	/**
	 * @return Root node of the regular expression tree
	 */
	const FormalRTEElement & getRTE ( ) const;

	/**
	 * @return Root node of the regular expression tree
	 */
	FormalRTEElement & getRTE ( );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRTE ( const FormalRTEElement & regExp );

	/**
	 * Sets the root node of the regular expression tree
	 * @param regExp root node to set
	 */
	void setRTE ( FormalRTEElement && regExp );

	/**
	 * Prints XML representation of the RTE to the output stream.
	 * @param out output stream to which print the RTE
	 * @param regexp RTE to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const FormalRTE & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "FormalRTE";

		return xmlTagName;
	}

	static FormalRTE parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace rte */

namespace std {

template < >
class ComponentConstraint< rte::FormalRTE, alphabet::RankedSymbol, rte::GeneralAlphabet > {
public:
	static bool used ( const rte::FormalRTE & rte, const alphabet::RankedSymbol & symbol ) {
		return rte.getRTE ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE & rte, const alphabet::RankedSymbol & symbol ) {
		if ( rte.accessComponent < rte::ConstantAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + ( std::string ) symbol + "cannot be in general alphabet since it is already in constant alphabet" );
	}
};

template < >
class ComponentConstraint< rte::FormalRTE, alphabet::RankedSymbol, rte::ConstantAlphabet > {
public:
	static bool used ( const rte::FormalRTE & rte, const alphabet::RankedSymbol & symbol ) {
		return rte.getRTE ( ).testSymbol ( symbol );
	}

	static bool available ( const rte::FormalRTE &, const alphabet::RankedSymbol & ) {
		return true;
	}

	static void valid ( const rte::FormalRTE & rte, const alphabet::RankedSymbol & symbol ) {
		if ( rte.accessComponent < rte::GeneralAlphabet > ( ).get ( ).count ( symbol ) )
			throw ::exception::CommonException ( "Symbol " + ( std::string ) symbol + "cannot be in constant alphabet since it is already in general alphabet" );
	}
};

} /* namespace std */

#endif /* FORMAL_RTE_H_ */
