/*
 * LinearStringTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "LinearStringTerminatingSymbol.h"
#include "LinearString.h"
#include <exception/CommonException.h>

#include <sstream>
#include <algorithm>

#include <sax/FromXMLParserHelper.h>
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include "String.h"
#include <object/Object.h>
#include <XmlApi.hpp>
#include <cast/CastApi.hpp>

namespace string {

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data ) : std::Components < LinearStringTerminatingSymbol, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( terminatingSymbol ) ) ) {
	setContent ( std::move ( data ) );
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol ) : LinearStringTerminatingSymbol ( std::set < alphabet::Symbol > { terminatingSymbol }, terminatingSymbol, std::vector < alphabet::Symbol > ( ) ) {
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data ) : LinearStringTerminatingSymbol ( std::set < alphabet::Symbol > ( data.begin ( ), data.end ( ) ) + std::set < alphabet::Symbol > { terminatingSymbol }, terminatingSymbol, data ) {
}

LinearStringTerminatingSymbol::LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, const LinearString & string ) : LinearStringTerminatingSymbol ( string.getAlphabet( ), std::move ( terminatingSymbol ), string.getContent ( ) ) {
}

StringBase * LinearStringTerminatingSymbol::clone ( ) const {
	return new LinearStringTerminatingSymbol ( * this );
}

StringBase * LinearStringTerminatingSymbol::plunder ( ) && {
	return new LinearStringTerminatingSymbol ( std::move ( * this ) );
}

void LinearStringTerminatingSymbol::appendSymbol ( alphabet::Symbol symbol ) {
	if ( getAlphabet().count ( symbol ) == 0 )
		throw exception::CommonException ( "Input symbol \"" + ( std::string ) symbol + "\" not in the alphabet." );

	m_Data.push_back ( std::move ( symbol ) );
}

const std::vector < alphabet::Symbol > & LinearStringTerminatingSymbol::getContent ( ) const {
	return this->m_Data;
}

void LinearStringTerminatingSymbol::setContent ( std::vector < alphabet::Symbol > data ) {
	std::set < alphabet::Symbol > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < alphabet::Symbol > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw exception::CommonException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

bool LinearStringTerminatingSymbol::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

int LinearStringTerminatingSymbol::compare ( const LinearStringTerminatingSymbol & other ) const {
	auto first = std::tie ( m_Data, getAlphabet() );
	auto second = std::tie ( other.m_Data, other.getAlphabet() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void LinearStringTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(LinearStringTerminatingSymbol ";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

LinearStringTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const alphabet::Symbol & symbol : this->m_Data )
		ss << ( std::string ) symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

LinearStringTerminatingSymbol LinearStringTerminatingSymbol::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LinearStringTerminatingSymbol::getXmlTagName() );
	std::set < alphabet::Symbol > alphabet = StringFromXMLParser::parseAlphabet ( input );
	alphabet::Symbol terminatingSymbol = alib::xmlApi < alphabet::Symbol >::parse ( input );
	std::vector < alphabet::Symbol > content = StringFromXMLParser::parseContent ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LinearStringTerminatingSymbol::getXmlTagName() );

	return LinearStringTerminatingSymbol ( alphabet, terminatingSymbol, content );
}

void LinearStringTerminatingSymbol::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LinearStringTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	StringToXMLComposer::compose ( out, getAlphabet() );
	alib::xmlApi < alphabet::Symbol >::compose ( out, getTerminatingSymbol ( ) );
	StringToXMLComposer::compose ( out, m_Data );
	out.emplace_back ( LinearStringTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

} /* namespace string */

namespace alib {

auto linearStringTerminatingSymbolParserRegister = xmlApi < string::String >::ParserRegister < string::LinearStringTerminatingSymbol > ();
auto linearStringTerminatingSymbolParserRegister2 = xmlApi < alib::Object >::ParserRegister < string::LinearStringTerminatingSymbol > ();

} /* namespace alib */
