/*
 * LinearStringTerminatingSymbol.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef LINEAR_STRING_TERMINATING_SYMBOL_H_
#define LINEAR_STRING_TERMINATING_SYMBOL_H_

#include "StringBase.h"
#include <iostream>
#include <set>
#include <vector>

#include "../alphabet/Symbol.h"
#include <core/components.hpp>

namespace string {

class LinearString;
class GeneralAlphabet;
class TerminatingSymbol;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class LinearStringTerminatingSymbol : public StringBase, public std::Components < LinearStringTerminatingSymbol, alphabet::Symbol, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > {
	std::vector < alphabet::Symbol > m_Data;

public:
	explicit LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol );
	explicit LinearStringTerminatingSymbol ( std::set < alphabet::Symbol > alphabet, alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data );
	explicit LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, std::vector < alphabet::Symbol > data );
	explicit LinearStringTerminatingSymbol ( alphabet::Symbol terminatingSymbol, const LinearString & string );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < alphabet::Symbol > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < alphabet::Symbol > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	const alphabet::Symbol & getTerminatingSymbol ( ) const {
		return accessElement < TerminatingSymbol > ( ).get ( );
	}

	/**
	 * @param element to append
	 */
	void appendSymbol ( alphabet::Symbol symbol );

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < alphabet::Symbol > & getContent ( ) const;

	void setContent ( std::vector < alphabet::Symbol > data );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LinearStringTerminatingSymbol & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LinearStringTerminatingSymbol";

		return xmlTagName;
	}

	static LinearStringTerminatingSymbol parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
};

} /* namespace string */

namespace std {

template < >
class ComponentConstraint< ::string::LinearStringTerminatingSymbol, alphabet::Symbol, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::LinearStringTerminatingSymbol & string, const alphabet::Symbol & symbol ) {
		const std::vector<alphabet::Symbol>& content = string.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::LinearStringTerminatingSymbol &, const alphabet::Symbol & ) {
		return true;
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const alphabet::Symbol & ) {
	}
};

template < >
class ElementConstraint< ::string::LinearStringTerminatingSymbol, alphabet::Symbol, ::string::TerminatingSymbol > {
public:
	static bool available ( const ::string::LinearStringTerminatingSymbol & string, const alphabet::Symbol & symbol ) {
		return string.getAlphabet ( ).count ( symbol );
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const alphabet::Symbol & ) {
	}
};

} /* namespace std */

#endif /* LINEAR_STRING_TERMINATING_SYMBOL_H_ */
