/*
 * RankedNode.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Stepan Plachy
 */

#include "RankedNode.h"
#include "RankedTree.h"
#include "../TreeException.h"
#include "../unranked/UnrankedNode.h"
#include <sstream>
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"

namespace tree {

RankedNode::RankedNode ( alphabet::RankedSymbol symbol, std::vector < std::smart_ptr < RankedNode > > children ) : symbol ( std::move ( symbol ) ), children ( std::move ( children ) ), parent ( NULL ), alphabet ( NULL ) {
	if ( this->children.size ( ) != this->symbol.getRank ( ).getData ( ) ) throw TreeException ( "Number of children doesn't match the rank of the symbol" );

	for ( std::smart_ptr < RankedNode > & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

RankedNode::RankedNode ( const RankedNode & other ) : symbol ( other.symbol ), children ( other.children ), parent ( NULL ), alphabet ( NULL ) {
	for ( std::smart_ptr < RankedNode > & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

RankedNode::RankedNode ( RankedNode && other ) noexcept : symbol ( std::move ( other.symbol ) ), children ( std::move ( other.children ) ), parent ( NULL ), alphabet ( NULL ) {
	for ( std::smart_ptr < RankedNode > & element : this->children )
		element->parent = this;

	this->attachAlphabet ( NULL );
}

RankedNode & RankedNode::operator =( const RankedNode & other ) {
	if ( this == & other )
		return * this;

	* this = RankedNode ( other );

	return * this;
}

RankedNode & RankedNode::operator =( RankedNode && other ) noexcept {
	std::swap ( this->symbol, other.symbol );
	std::swap ( this->children, other.children );
	std::swap ( this->alphabet, other.alphabet ); // this->alphabet is stored within other.alphabet and it is reattached on the next line

	for ( auto & element : this->children )
		element->parent = this;

	this->attachAlphabet ( other.alphabet );

	return * this;
}

RankedNode::~RankedNode ( ) noexcept {
}

const std::vector < std::smart_ptr < const RankedNode > > & RankedNode::getChildren ( ) const {
	return reinterpret_cast < const std::vector < std::smart_ptr < const RankedNode > > & > ( children );
}

const std::vector < std::smart_ptr < RankedNode > > & RankedNode::getChildren ( ) {
	return children;
}

RankedNode * RankedNode::getParent ( ) {
	return parent;
}

const RankedNode * RankedNode::getParent ( ) const {
	return parent;
}

void RankedNode::setSymbol ( alphabet::RankedSymbol symbol ) {
	if ( symbol.getRank ( ) != this->symbol.getRank ( ) ) throw TreeException ( "Number of children doesn't match the rank of the symbol" );

	this->symbol = std::move ( symbol );

	if ( ( this->alphabet != NULL ) && ( this->alphabet->count ( this->symbol ) == 0 ) ) throw TreeException ( "Symbol is not in the alphabet" );
}

void RankedNode::swap ( RankedNode & other ) {
	const std::set < alphabet::RankedSymbol > * thisParentTree = this->alphabet;
	const std::set < alphabet::RankedSymbol > * otherParentTree = other.alphabet;

	RankedNode tmp = std::move ( other );

	other  = std::move ( * this );
	* this = std::move ( tmp );

	this->attachAlphabet ( thisParentTree );
	other.attachAlphabet ( otherParentTree );
}

UnrankedNode RankedNode::asUnranked ( ) const {
	std::vector < std::smart_ptr < UnrankedNode > > unrankedChildren;

	for ( const std::smart_ptr < const RankedNode > & child : this->getChildren() )
		unrankedChildren.emplace_back ( std::make_smart < UnrankedNode > ( child->asUnranked ( ) ) );

	return UnrankedNode ( this->symbol.getSymbol ( ), std::move ( unrankedChildren ) );
}

int RankedNode::compare ( const RankedNode & other ) const {
	auto first = std::tie ( symbol, getChildren() );
	auto second = std::tie ( other.symbol, other.getChildren() );

	std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

void RankedNode::operator >>( std::ostream & out ) const {
	out << "(RankedNode " << " symbol = " << this->symbol << " children = " << this->children << "})";
}

std::ostream & operator <<( std::ostream & out, const RankedNode & node ) {
	node >> out;
	return out;
}

bool RankedNode::testSymbol ( const alphabet::RankedSymbol & symbol ) const {
	if ( symbol == this->symbol ) return true;

	for ( const std::smart_ptr < const RankedNode > & child : this->getChildren() )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

bool RankedNode::attachAlphabet ( const std::set < alphabet::RankedSymbol > * alphabet ) {
	if ( this->alphabet == alphabet ) return true;

	this->alphabet = alphabet;

	for ( std::smart_ptr < RankedNode > & child : this->children )
		if ( !child->attachAlphabet ( alphabet ) ) return false;

	if ( ( this->alphabet != NULL ) && ( this->alphabet->count ( this->symbol ) == 0 ) ) return false;

	return true;
}

std::set < alphabet::RankedSymbol > RankedNode::computeMinimalAlphabet ( ) const {
	std::set < alphabet::RankedSymbol > res;
	computeMinimalAlphabet ( res );
	return res;
}

void RankedNode::computeMinimalAlphabet ( std::set < alphabet::RankedSymbol > & alphabet ) const {
	alphabet.insert ( this->symbol );

	for ( const std::smart_ptr < const RankedNode > & child : this->getChildren() )
		child->computeMinimalAlphabet ( alphabet );
}

const alphabet::RankedSymbol & RankedNode::getSymbol ( ) const {
	return this->symbol;
}

RankedNode::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

void RankedNode::nicePrint ( std::ostream & os, const std::string & prefix, const bool last ) const {
	os << prefix;

	std::string nextPrefix ( prefix );

	if ( last ) {
		os << "\\-";
		nextPrefix += "  ";
	} else {
		os << "|-";
		nextPrefix += "| ";
	}

	os << ( std::string ) symbol.getSymbol ( ) << "(" << symbol.getRank ( ).getData ( ) << ")" << std::endl;

	for ( unsigned int i = 0; i < children.size ( ); ++i ) {
		os << nextPrefix << "|" << std::endl;
		children[i]->nicePrint ( os, nextPrefix, i == children.size ( ) - 1 );
	}
}

} /* namespace tree */

namespace alib {

tree::RankedNode xmlApi < tree::RankedNode >::parse ( std::deque < sax::Token >::iterator & input ) {
	return tree::TreeFromXMLParser::parseRankedNode ( input );
}

bool xmlApi < tree::RankedNode >::first ( const std::deque < sax::Token >::const_iterator & ) {
	throw tree::TreeException ( "Unimplemented" );
}

void xmlApi < tree::RankedNode >::compose ( std::deque < sax::Token > & output, const tree::RankedNode & data ) {
	tree::TreeToXMLComposer::composeNode ( output, data );
}

} /* namespace alib */
