#include "TreeTest.h"

#include "sax/SaxParseInterface.h"
#include "sax/SaxComposeInterface.h"

#include "tree/ranked/RankedTree.h"
#include "tree/unranked/UnrankedTree.h"
#include "tree/ranked/PrefixRankedTree.h"
#include "tree/ranked/PrefixRankedBarTree.h"
#include "tree/ranked/RankedPattern.h"

#include "tree/ranked/RankedNode.h"
#include "tree/unranked/UnrankedNode.h"

#include "tree/TreeException.h"

#include "factory/XmlDataFactory.hpp"

#include "alphabet/RankedSymbol.h"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/BarSymbol.h"

#define CPPUNIT_IMPLY(x, y) CPPUNIT_ASSERT(!(x) || (y))

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( TreeTest, "tree" );
CPPUNIT_TEST_SUITE_REGISTRATION( TreeTest );

void TreeTest::setUp() {
}

void TreeTest::tearDown() {
}

tree::RankedNode prefixToNode(const std::string & s, int & pos) {
	char c = s[pos];
	int r = s[pos+1] - '0';
	pos += 2;
	std::vector<std::smart_ptr<tree::RankedNode>> children;
	for (int i = 0; i < r; i++) {
		children.emplace_back(std::make_smart<tree::RankedNode>(prefixToNode(s, pos)));
	}
	return tree::RankedNode(alphabet::RankedSymbol(c, r), std::move(children));
}

void TreeTest::testRankedTreeParser() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)
*/
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	tree::RankedNode node3(c, {});
	tree::RankedNode node4(c, {});
	tree::RankedNode node2(b, {std::make_smart<tree::RankedNode>(std::move(node3))});
	tree::RankedNode node1(a, {std::make_smart<tree::RankedNode>(std::move(node2)), std::make_smart<tree::RankedNode>(std::move(node4))});

	tree::RankedTree tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getRoot().nicePrint(std::cout);
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::RankedTree tree2 = alib::XmlDataFactory::fromTokens<tree::RankedTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getRoot().nicePrint(std::cout);
	}
/*
	std::string s = "a2a2a2b1a2c0c0c0c0c0";
	int itmp = 0;
	tree::RankedNode * node = prefixToNode(s, itmp);
	std::set<alphabet::RankedSymbol> al;
	for (unsigned i = 0; i < s.length(); i += 2) al.insert(alphabet::RankedSymbol(s[i], (int) (s[i+1] - '0')));
	tree::RankedTree t(al, std::move(*node));
	alib::XmlDataFactory::toStdout(t);
*/
}

void TreeTest::testRankedTreeCompare() {
/*
	L=(a(2), b(1), c(0))

	\-1.a(2)
	  |
	  |-2.b(1)
	  | |
	  | \-3.c(0)
	  |
	  \-4.c(0)

	\-5.a(2)
	  |
	  |-6.c(0)
	  |
	  \-7.b(1)
	    |
	    \-8.c(0)
*/
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	tree::RankedNode node3(c, {});
	tree::RankedNode node4(c, {});
	tree::RankedNode node2(b, {std::make_smart<tree::RankedNode>(std::move(node3))});
	tree::RankedNode node1(a, {std::make_smart<tree::RankedNode>(std::move(node2)), std::make_smart<tree::RankedNode>(std::move(node4))});

	tree::RankedNode node6(c, {});
	tree::RankedNode node8(c, {});
	tree::RankedNode node7(b, {std::make_smart<tree::RankedNode>(std::move(node8))});
	tree::RankedNode node5(a, {std::make_smart<tree::RankedNode>(std::move(node6)), std::make_smart<tree::RankedNode>(std::move(node7))});

	tree::RankedTree tree1(alphabet, std::move(node1));
	tree::RankedTree tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testRankedTreeSymbolValidityCheck() {
	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);
	const alphabet::RankedSymbol d ('d', 0);
	const alphabet::RankedSymbol e ('e', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c, e};

	tree::RankedNode node3(d, {});
	tree::RankedNode node4(c, {});
	tree::RankedNode node2(b, {std::make_smart<tree::RankedNode>(std::move(node3))});
	tree::RankedNode node1(a, {std::make_smart<tree::RankedNode>(std::move(node2)), std::make_smart<tree::RankedNode>(std::move(node4))});

	CPPUNIT_ASSERT_THROW(tree::RankedTree(alphabet, std::move(node1)), tree::TreeException);

	tree::RankedNode node6(e, {});
	tree::RankedNode node8(c, {});
	tree::RankedNode node7(b, {std::make_smart<tree::RankedNode>(std::move(node8))});
	tree::RankedNode node5(a, {std::make_smart<tree::RankedNode>(std::move(node6)), std::make_smart<tree::RankedNode>(std::move(node7))});

	tree::RankedTree * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::RankedTree(alphabet, std::move(node5)));

	CPPUNIT_ASSERT_NO_THROW(tree -> getRoot().getChildren()[0] -> setSymbol(c));
	CPPUNIT_ASSERT_THROW(tree -> getRoot().getChildren()[1] -> setSymbol(d), tree::TreeException);
	CPPUNIT_ASSERT_THROW(tree -> getRoot().getChildren()[0] -> setSymbol(a), tree::TreeException);

	delete tree;
}

void TreeTest::testRankedTreeSubtreeSwitch() {
	alphabet::RankedSymbol a ('a', 2);
	alphabet::RankedSymbol b ('b', 1);
	alphabet::RankedSymbol c ('c', 0);
	alphabet::RankedSymbol d ('d', 0);

	const std::set<alphabet::RankedSymbol> alphabet1 {a, b, c, d};
	const std::set<alphabet::RankedSymbol> alphabet2 {a, b, c, d};

	tree::RankedNode node3(c, {});
	tree::RankedNode node4(c, {});
	tree::RankedNode node2(b, {std::make_smart<tree::RankedNode>(std::move(node3))});
	tree::RankedNode node1(a, {std::make_smart<tree::RankedNode>(std::move(node2)), std::make_smart<tree::RankedNode>(std::move(node4))});

	tree::RankedNode node6(d, {});
	tree::RankedNode node8(c, {});
	tree::RankedNode node7(b, {std::make_smart<tree::RankedNode>(std::move(node8))});
	tree::RankedNode node5(a, {std::make_smart<tree::RankedNode>(std::move(node6)), std::make_smart<tree::RankedNode>(std::move(node7))});

	tree::RankedTree tree1(alphabet1, std::move(node1));
	tree::RankedTree tree1Copy (tree1);
	tree::RankedTree tree2(alphabet2, std::move(node5));
	tree::RankedTree tree2Copy (tree2);

	tree1.getRoot().getChildren()[0] -> getChildren()[0] -> swap(*(tree2.getRoot().getChildren()[1]));
	tree1.getRoot().getChildren()[0] -> getChildren()[0] -> swap(*(tree2.getRoot().getChildren()[1]));

	CPPUNIT_ASSERT(tree1 == tree1Copy);
	CPPUNIT_ASSERT(tree2 == tree2Copy);
	tree1.getRoot().getChildren()[1] -> swap(*(tree2.getRoot().getChildren()[0]));

	tree::RankedTree tree3(tree1Copy);
	tree::RankedTree tree4(tree2Copy);

	tree4.getRoot().getChildren()[0] -> setSymbol(c);

	tree3.getRoot().nicePrint(std::cout);
	std::cout << std::endl;
	tree4.getRoot().nicePrint(std::cout);

	tree3.getRoot().swap(tree4.getRoot());

	CPPUNIT_ASSERT(tree3 < tree2Copy);
	CPPUNIT_ASSERT(tree4 == tree1Copy);
}

void TreeTest::testUnrankedTreeParser() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c
*/
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');

	const std::set<alphabet::Symbol> alphabet {a, b, c};

	tree::UnrankedNode node3(c, {});
	tree::UnrankedNode node4(c, {});
	tree::UnrankedNode node2(b, {std::make_smart<tree::UnrankedNode>(std::move(node3))});
	tree::UnrankedNode node1(a, {std::make_smart<tree::UnrankedNode>(std::move(node2)), std::make_smart<tree::UnrankedNode>(std::move(node4))});

	tree::UnrankedTree tree(alphabet, std::move(node1));

	CPPUNIT_ASSERT( tree == tree );
	tree.getRoot().nicePrint(std::cout);
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);
		std::cout << std::endl << tmp << std::endl << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::UnrankedTree tree2 = alib::XmlDataFactory::fromTokens<tree::UnrankedTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree == tree2 );
		std::cout << std::endl;
		tree2.getRoot().nicePrint(std::cout);
	}
}

void TreeTest::testUnrankedTreeCompare() {
/*
	L=(a, b, c)

	\-1.a
	  |
	  |-2.b
	  | |
	  | \-3.c
	  |
	  \-4.c

	\-5.a
	  |
	  |-6.c
	  |
	  \-7.b
	    |
	    \-8.c
*/
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');

	const std::set<alphabet::Symbol> alphabet {a, b, c};

	tree::UnrankedNode node3(c, {});
	tree::UnrankedNode node4(c, {});
	tree::UnrankedNode node2(b, {std::make_smart<tree::UnrankedNode>(std::move(node3))});
	tree::UnrankedNode node1(a, {std::make_smart<tree::UnrankedNode>(std::move(node2)), std::make_smart<tree::UnrankedNode>(std::move(node4))});

	tree::UnrankedNode node6(c, {});
	tree::UnrankedNode node8(c, {});
	tree::UnrankedNode node7(b, {std::make_smart<tree::UnrankedNode>(std::move(node8))});
	tree::UnrankedNode node5(a, {std::make_smart<tree::UnrankedNode>(std::move(node6)), std::make_smart<tree::UnrankedNode>(std::move(node7))});

	tree::UnrankedTree tree1(alphabet, std::move(node1));
	tree::UnrankedTree tree2(alphabet, std::move(node5));

	CPPUNIT_ASSERT( tree1 != tree2 );
	CPPUNIT_ASSERT( tree1 < tree2 );
	CPPUNIT_ASSERT( tree2 > tree1 );
	CPPUNIT_ASSERT( !(tree1 > tree2) );
}

void TreeTest::testUnrankedTreeSymbolValidityCheck() {
	const alphabet::Symbol a = alphabet::symbolFrom('a');
	const alphabet::Symbol b = alphabet::symbolFrom('b');
	const alphabet::Symbol c = alphabet::symbolFrom('c');
	const alphabet::Symbol d = alphabet::symbolFrom('d');
	const alphabet::Symbol e = alphabet::symbolFrom('e');

	const std::set<alphabet::Symbol> alphabet {a, b, c, e};

	tree::UnrankedNode node3(d, {});
	tree::UnrankedNode node4(c, {});
	tree::UnrankedNode node2(b, {std::make_smart<tree::UnrankedNode>(std::move(node3))});
	tree::UnrankedNode node1(a, {std::make_smart<tree::UnrankedNode>(std::move(node2)), std::make_smart<tree::UnrankedNode>(std::move(node4))});

	CPPUNIT_ASSERT_THROW(tree::UnrankedTree(alphabet, std::move(node1)), tree::TreeException);

	tree::UnrankedNode node6(e, {});
	tree::UnrankedNode node8(c, {});
	tree::UnrankedNode node7(b, {std::make_smart<tree::UnrankedNode>(std::move(node8))});
	tree::UnrankedNode node5(a, {std::make_smart<tree::UnrankedNode>(std::move(node6)), std::make_smart<tree::UnrankedNode>(std::move(node7))});

	tree::UnrankedTree * tree = NULL;
	CPPUNIT_ASSERT_NO_THROW(tree = new tree::UnrankedTree(alphabet, std::move(node5)));

	CPPUNIT_ASSERT_NO_THROW((*tree -> getRoot().getChildren().begin()) -> setSymbol(c));
	CPPUNIT_ASSERT_THROW((*tree -> getRoot().getChildren().begin()) -> setSymbol(d), tree::TreeException);
	CPPUNIT_ASSERT_NO_THROW((*(++tree -> getRoot().getChildren().begin())) -> setSymbol(a));
	delete tree;
}

void TreeTest::testUnrankedTreeSubtreeSwitch() {
	alphabet::Symbol a = alphabet::symbolFrom('a');
	alphabet::Symbol b = alphabet::symbolFrom('b');
	alphabet::Symbol c = alphabet::symbolFrom('c');
	alphabet::Symbol d = alphabet::symbolFrom('d');

	const std::set<alphabet::Symbol> alphabet1 {a, b, c, d};
	const std::set<alphabet::Symbol> alphabet2 {a, b, c, d};

	tree::UnrankedNode node3(c, {});
	tree::UnrankedNode node4(c, {});
	tree::UnrankedNode node2(b, {std::make_smart<tree::UnrankedNode>(std::move(node3))});
	tree::UnrankedNode node1(a, {std::make_smart<tree::UnrankedNode>(std::move(node2)), std::make_smart<tree::UnrankedNode>(std::move(node4))});

	tree::UnrankedNode node6(d, {});
	tree::UnrankedNode node8(c, {});
	tree::UnrankedNode node7(b, {std::make_smart<tree::UnrankedNode>(std::move(node8))});
	tree::UnrankedNode node5(a, {std::make_smart<tree::UnrankedNode>(std::move(node6)), std::make_smart<tree::UnrankedNode>(std::move(node7))});

	tree::UnrankedTree tree1(alphabet1, std::move(node1));
	tree::UnrankedTree tree1Copy (tree1);
	tree::UnrankedTree tree2(alphabet2, std::move(node5));
	tree::UnrankedTree tree2Copy (tree2);

	(*(*tree1.getRoot().getChildren().begin()) -> getChildren().begin()) -> swap(**(++tree2.getRoot().getChildren().begin()));
	(*(*tree1.getRoot().getChildren().begin()) -> getChildren().begin()) -> swap(**(++tree2.getRoot().getChildren().begin()));

	CPPUNIT_ASSERT(tree1 == tree1Copy);
	CPPUNIT_ASSERT(tree2 == tree2Copy);
	(*(++tree1.getRoot().getChildren().begin())) -> swap(**tree2.getRoot().getChildren().begin());

	tree::UnrankedTree tree3(tree1Copy);
	tree::UnrankedTree tree4(tree2Copy);

	(*tree4.getRoot().getChildren().begin()) -> setSymbol(c);

	tree3.getRoot().nicePrint(std::cout);
	std::cout << std::endl;
	tree4.getRoot().nicePrint(std::cout);

	tree3.getRoot().swap(tree4.getRoot());

	CPPUNIT_ASSERT(tree3 < tree2Copy);
	CPPUNIT_ASSERT(tree4 == tree1Copy);
}

void TreeTest::testPrefixRankedBarParser() {
	const alphabet::Symbol bar{alphabet::BarSymbol{}};

	const alphabet::RankedSymbol a ('a', 2);
	const alphabet::RankedSymbol b ('b', 1);
	const alphabet::RankedSymbol c ('c', 0);

	const std::set<alphabet::RankedSymbol> alphabet {a, b, c};

	tree::RankedNode node3(c, {});
	tree::RankedNode node4(c, {});
	tree::RankedNode node2(b, {std::make_smart<tree::RankedNode>(std::move(node3))});
	tree::RankedNode node1(a, {std::make_smart<tree::RankedNode>(std::move(node2)), std::make_smart<tree::RankedNode>(std::move(node4))});

	tree::RankedTree tree(alphabet, std::move(node1));
	tree::PrefixRankedBarTree tree2(bar, tree);

	CPPUNIT_ASSERT( tree2 == tree2 );
	{
		std::deque<sax::Token> tokens = alib::XmlDataFactory::toTokens(tree2);
		std::string tmp;
		sax::SaxComposeInterface::printMemory(tmp, tokens);

		std::cout << tmp << std::endl << std::endl;

		std::deque<sax::Token> tokens2;
		sax::SaxParseInterface::parseMemory(tmp, tokens2);
		tree::PrefixRankedBarTree tree3 = alib::XmlDataFactory::fromTokens<tree::PrefixRankedBarTree>(std::move(tokens2));

		CPPUNIT_ASSERT( tree2 == tree3 );
		std::cout << std::endl;
	}
}

