/*
 * AllEpsilonClosure.cpp
 *
 *  Created on: 12. 4. 2015
 *	  Author: Jan Travnicek
 */

#include "AllEpsilonClosure.h"

#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/NFA.h>
#include <automaton/FSM/DFA.h>

#include <set>
#include <map>
#include <iostream>

#include <regexp/properties/RegExpEpsilon.h>


namespace automaton {

namespace properties {

namespace efficient {

void process(const std::map<automaton::State, std::set<automaton::State>>& epsilonTransitions, const automaton::State * state, std::map<automaton::State, std::set<automaton::State>>& closures, std::set<automaton::State>& visited, std::set<automaton::State>& nonvisited) {
	if(visited.count(*state)) return;

	state = &*visited.insert(*state).first;
	nonvisited.erase(*state);

	auto tos = epsilonTransitions.find(*state);
	if(tos == epsilonTransitions.end()) return;

	for(const automaton::State& to : tos->second) {
		process(epsilonTransitions, &to, closures, visited, nonvisited);
		closures[*state].insert(closures[to].begin(), closures[to].end());
	}
}

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::EpsilonNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;

	for(const automaton::State& state : fsm.getStates())
		res[state].insert(state);

	std::set<automaton::State> visited;
	std::set<automaton::State> nonvisited = fsm.getStates();

	std::map<automaton::State, std::set<automaton::State>> epsilonTransitions = fsm.getEpsilonTransitions();

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureEpsilonNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::EpsilonNFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::MultiInitialStateNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureMultiInitialStateNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::MultiInitialStateNFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::NFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::NFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::DFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> closure;
	for(const automaton::State& state : fsm.getStates())
		closure[state].insert(state);
	return closure;
}

auto AllEpsilonClosureDFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::DFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::ExtendedNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;

	for(const automaton::State& state : fsm.getStates())
		res[state].insert(state);

	std::set<automaton::State> visited;
	std::set<automaton::State> nonvisited = fsm.getStates();

	std::map<automaton::State, std::set<automaton::State>> epsilonTransitions;
	for(const std::pair<const std::pair<State, regexp::RegExp>, std::set<State> >& transition : fsm.getTransitions() )
		if( regexp::properties::RegExpEpsilon::languageContainsEpsilon( transition.first.second ) )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureExtendedNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::ExtendedNFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure( const automaton::CompactNFA & fsm) {
	std::map<automaton::State, std::set<automaton::State>> res;

	for(const automaton::State& state : fsm.getStates())
		res[state].insert(state);

	std::set<automaton::State> visited;
	std::set<automaton::State> nonvisited = fsm.getStates();

	std::map<automaton::State, std::set<automaton::State>> epsilonTransitions;
	for(const std::pair<const std::pair<State, string::LinearString>, std::set<State> >& transition : fsm.getTransitions() )
		if( transition.first.second.getContent().size() == 0 )
			epsilonTransitions[transition.first.first].insert(transition.second.begin(), transition.second.end());

	while(nonvisited.size()) {
		process(epsilonTransitions, &*nonvisited.begin(), res, visited, nonvisited);
	}

	// to correctly handle epsilon loops (swap visited and nonvisited)

	while(visited.size()) {
		process(epsilonTransitions, &*visited.begin(), res, nonvisited, visited);
	}

	return res;
}

auto AllEpsilonClosureCompactNFA = AllEpsilonClosure::RegistratorWrapper<std::map<automaton::State, std::set<automaton::State>>, automaton::CompactNFA>(AllEpsilonClosure::allEpsilonClosure);

std::map<automaton::State, std::set<automaton::State>> AllEpsilonClosure::allEpsilonClosure(const Automaton& automaton) {
	return dispatch(automaton.getData());
}

} /* namespace efficient */

} /* namespace properties */

} /* namespace automaton */

