/*
 * Author: Radovan Cerveny
 */

#include "exception/CommonException.h"
#include "sax/SaxParseInterface.h"
#include <global/GlobalData.h>
#include <tclap/CmdLine.h>

#include <measure>
#include <measurepp>

int main ( int argc, char * * argv ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Measurement processor binary", ' ', "0.01" );

		std::vector < std::string > allowedOutput;
		allowedOutput.push_back ( "xml" );
		allowedOutput.push_back ( "stats" );
		allowedOutput.push_back ( "csv" );
		allowedOutput.push_back ( "html" );
		TCLAP::ValuesConstraint < std::string > allowedOutputValues ( allowedOutput );

		TCLAP::ValueArg < std::string > outputFormat ( "o", "output", "Output format", false, "xml", & allowedOutputValues );
		cmd.add ( outputFormat );

		std::vector < std::string > allowedEngine;
		allowedEngine.push_back ( "time" );
		allowedEngine.push_back ( "memory" );
		allowedEngine.push_back ( "counter" );
		TCLAP::ValuesConstraint < std::string > allowedEngineValues ( allowedEngine );

		TCLAP::ValueArg < std::string > engine ( "e", "engine", "Engine type", false, "time", & allowedEngineValues );
		cmd.add ( engine );

		TCLAP::MultiArg < std::string > engineAttr ( "", "engineAttr", "Engine attribute", false, "string" );
		cmd.add ( engineAttr );

		TCLAP::MultiArg < std::string > filterByFrameType ( "", "filterFrameType", "Filter by frame type", false, "string" );
		cmd.add ( filterByFrameType );

		TCLAP::MultiArg < std::string > filterByFrameName ( "", "filterFrameName", "Filter by frame name", false, "string" );
		cmd.add ( filterByFrameName );

		TCLAP::MultiArg < std::string > filterByCommand ( "", "filterCommand", "Filter by command", false, "string" );
		cmd.add ( filterByCommand );

		TCLAP::SwitchArg aggregate ( "", "aggregate", "Aggregate results over the same combination of inputs", false );
		cmd.add ( aggregate );

		TCLAP::UnlabeledValueArg < std::string > input ( "input", "provisioning results file", false, "-", "file" );
		cmd.add ( input );

		TCLAP::SwitchArg verbose(		"v",	"verbose",	"Be verbose",			false);
		cmd.add( verbose );

		cmd.parse(argc, argv);

		if(verbose.isSet())
			common::GlobalData::verbose = true;

		std::deque < sax::Token > tokens;

		if ( input.getValue ( ) == "-" )
			sax::SaxParseInterface::parseStdin ( tokens );
		else
			sax::SaxParseInterface::parseFile ( input.getValue ( ), tokens );

		auto results = measurements::MeasurementProvisionerResultsXml::parse ( tokens );

		measurements::MeasurementProcessorFilter filter;

		for ( const std::string & f : filterByFrameType.getValue ( ) )
			filter.add ( measurements::MeasurementProcessorFilter::FilterType::FRAME_TYPE, f );

		for ( const std::string & f : filterByFrameName.getValue ( ) )
			filter.add ( measurements::MeasurementProcessorFilter::FilterType::FRAME_NAME, f );

		for ( const std::string & f : filterByCommand.getValue ( ) )
			filter.add ( measurements::MeasurementProcessorFilter::FilterType::COMMAND, f );

		auto processedResults = measurements::MeasurementProcessor::process ( results, filter );

		if ( aggregate.isSet ( ) )
			processedResults = measurements::MeasurementProcessor::aggregate ( processedResults );

		measurements::MeasurementProcessorOutput output;

		if ( outputFormat.getValue ( ) == "xml" )
			output.outputFormat = measurements::MeasurementProcessorOutput::OutputFormat::XML;
		else if ( outputFormat.getValue ( ) == "stats" )
			output.outputFormat = measurements::MeasurementProcessorOutput::OutputFormat::STATS;
		else if ( outputFormat.getValue ( ) == "csv" )
			output.outputFormat = measurements::MeasurementProcessorOutput::OutputFormat::CSV;
		else if ( outputFormat.getValue ( ) == "html" )
			output.outputFormat = measurements::MeasurementProcessorOutput::OutputFormat::HTML;

		if ( engine.getValue ( ) == "time" )
			output.outputEngine = measurements::MeasurementProcessorOutput::OutputEngine::TIME;
		else if ( engine.getValue ( ) == "memory" )
			output.outputEngine = measurements::MeasurementProcessorOutput::OutputEngine::MEMORY;
		else if ( engine.getValue ( ) == "counter" )
			output.outputEngine = measurements::MeasurementProcessorOutput::OutputEngine::COUNTER;

		if ( engineAttr.isSet ( ) )
			for ( const std::string & attr : engineAttr.getValue ( ) )
				output.outputEngineAttr.insert ( attr );

		std::cout << measurements::MeasurementProcessor::output ( processedResults, output ) << std::endl;

		return 0;

	} catch ( const exception::CommonException & exception ) {
		std::cerr << "CommonException caught: " << exception.getCause ( ) << std::endl;
		return 4;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
